from copy import deepcopy
from typing import Any, Dict, List

import mosaik_api_v3
from mosaik_api_v3.types import (
    CreateResult,
    InputData,
    Meta,
    ModelName,
    OutputData,
    OutputRequest,
    SimId,
    Time,
)

import mosaik
import mosaik.exceptions

META: Meta = {
    "api_version": "3.0",
    "type": "event-based",
    "extra_methods": ["get_dict"],
    "models": {
        "Dict": {
            "public": True,
            "any_inputs": True,
            "params": [],
            "attrs": [],
        },
    },
}


class OutputSimulator(mosaik_api_v3.Simulator):
    """
    This simulator takes the input it is given and writes it into a python dictionary
    where the keys are the timestamps of the input and the values are the inputs values.

    The dictionary can be retrieved using the :meth:`get_dict` method.
    """

    entities: Dict[str, Dict[Time, Any]]

    def __init__(self):
        super().__init__(META)
        self.entities = {}  # Maps EIDs to model instances/entities

    def init(self, sid: SimId, time_resolution: float = 1):
        return self.meta

    def create(
        self, num: int, model: ModelName, **model_params: Any
    ) -> List[CreateResult]:
        next_eid = len(self.entities)
        entities: List[CreateResult] = []
        for i in range(next_eid, next_eid + num):
            model_instance = {}
            eid = f"{model}-{i}"
            self.entities[eid] = model_instance
            entities.append({"eid": eid, "type": model})
        return entities

    def step(self, time: Time, inputs: InputData, max_advance: Time):
        for receiver, input in inputs.items():
            self.entities[receiver][time] = deepcopy(input)

    def get_data(self, outputs: OutputRequest) -> OutputData:
        raise mosaik.exceptions.ScenarioError(
            "This function is not supposed to be used in this simulator. "
            "Use this simulator for input data only."
        )

    def get_dict(self, eid: str) -> Dict[Time, Any]:
        """
        Returns the dict of the simulator entity specified by the ``eid``.

        :param eid: The entity id of the selected entity.
        :type eid: str

        :return: The dict of the entity.
        """
        return self.entities[eid]



"""
This module provides convenient access to all classes and functions required
to create scenarios and run simulations.

Currently, this is only :class:`mosaik.scenario.World`.

"""

from loguru import logger

from mosaik._version import version

# Saying "import X as X" (with X repeated) is the standard way of
# marking to linters and type checkers that something is re-exported.
from mosaik.async_scenario import AsyncWorld as AsyncWorld
from mosaik.scenario import SimConfig as SimConfig
from mosaik.scenario import World as World

__all__ = ["World", "AsyncWorld"]
__version__ = version

logger.disable(__name__)



"""
This module is responsible for performing the simulation of a scenario.
"""

from __future__ import annotations

import asyncio
from heapq import heappop
from math import ceil
from time import perf_counter
from typing import TYPE_CHECKING, Any, Coroutine, Dict, List, Optional

from loguru import logger
from mosaik_api_v3 import InputData, SimId, Time

from mosaik.exceptions import SimulationError
from mosaik.internal_util import merge_all, merge_existing
from mosaik.simmanager import FULL_ID, SimRunner
from mosaik.tiered_time import TieredTime

if TYPE_CHECKING:
    from mosaik.async_scenario import AsyncWorld


SENTINEL = object()


async def run(
    world: AsyncWorld,
    until: int,
    rt_factor: Optional[float] = None,
    rt_strict: bool = False,
    lazy_stepping: bool = True,
):
    """
    Run the simulation for a :class:`~mosaik.scenario.AsyncWorld` until
    the simulation time *until* has been reached.

    Return the final simulation time.

    See :meth:`mosaik.scenario.AsyncWorld.run()` for a detailed description of the
    *rt_factor* and *rt_strict* arguments.
    """
    world.until = until

    if rt_factor is not None and rt_factor <= 0:
        raise ValueError('"rt_factor" is %s but must be > 0"' % rt_factor)
    if rt_factor is not None:
        # Adjust rt_factor to the time_resolution:
        rt_factor *= world.time_resolution
    world.rt_factor = rt_factor

    setup_done_events: List[asyncio.Task[None]] = []
    for sim in world.sims.values():
        sim.tqdm.set_postfix_str("setup")
        # Send a setup_done event to all simulators
        setup_done_events.append(asyncio.create_task(sim.setup_done()))

    # Wait for all answers to be here
    await asyncio.gather(*setup_done_events)

    # Start simulator processes
    processes: List[asyncio.Task[None]] = []
    for sim in world.sims.values():
        process = asyncio.create_task(
            sim_process(world, sim, until, rt_factor, rt_strict, lazy_stepping),
            name=f"Runner for {sim.sid}",
        )

        sim.task = process
        processes.append(process)

    # Wait for all processes to be done
    await asyncio.gather(*processes)


async def sim_process(
    world: AsyncWorld,
    sim: SimRunner,
    until: int,
    rt_factor: Optional[float],
    rt_strict: bool,
    lazy_stepping: bool,
):
    """
    Coroutine running the simulator *sim*.
    """
    sim.started = True
    sim.rt_start = rt_start = perf_counter()

    try:
        advance_progress(sim, world)
        while await next_step_settled(sim, world):
            sim.tqdm.set_postfix_str("await input")
            await wait_for_dependencies(sim, lazy_stepping)
            sim.current_step = heappop(sim.next_steps)
            if sim.current_step != sim.progress.time:
                raise SimulationError(
                    f"Simulator {sim.sid} is trying to perform a step at time "
                    f"{sim.current_step}, but it has already progressed to time "
                    f"{sim.progress.time}."
                )
            if any(t >= world.max_loop_iterations for t in sim.current_step.tiers[1:]):
                raise SimulationError(
                    f"Simulator {sim.sid} has performed a sub-step more than "
                    f"{world.max_loop_iterations} times. (The complete now is "
                    f"{sim.current_step}.) This might indicate that you have run into "
                    "an infinite loop. If not, you can increase max_loop_iterations to "
                    "get rid of this warning."
                )
            input_data = get_input_data(world, sim)
            max_advance = get_max_advance(world, sim, until)
            await step(world, sim, input_data, max_advance)
            rt_check(rt_factor, rt_start, rt_strict, sim)
            await get_outputs(world, sim)
            sim.current_step = None
            trigger_successors(sim)
            # TODO: Reduce the number of sims that need to be advanced
            # (At least only to those that could potentially be
            # triggered by this step; maybe there's even a more clever
            # way.)
            for isim in world.sims.values():
                advance_progress(isim, world)
            world.sim_progress = get_progress(world.sims, until)
            world.tqdm.update(get_avg_progress(world.sims, until) - world.tqdm.n)
            if world.use_cache:
                prune_dataflow_cache(world)
        sim.tqdm.set_postfix_str("done")
    except ConnectionError as e:
        raise SimulationError('Simulator "%s" closed its connection.' % sim.sid, e)


async def next_step_settled(sim: SimRunner, world: AsyncWorld) -> bool:
    # When deciding when the next step will happen, we have two numbers
    # that approach each other: The earliest currently scheduled next
    # step (which might still go down) and the earliest potential next
    # step based on the rest of the simulation, which we call the
    # simulator's progress and which can only go up.
    # Once these two numbers meet, we know that that time is the next
    # step.
    # As a slight complication, we also need to watch out for the end
    # of the simulation. Once that is reached, we also return, albeit
    # without having found a next step.
    sim.tqdm.set_postfix_str("await step")
    while sim.progress.time.time < world.until:
        if sim.next_steps and sim.next_steps[0] == sim.progress.time:
            return True
        else:
            await_time = (
                sim.next_steps[0]
                if sim.next_steps
                else TieredTime(world.until) + sim.from_world_time
            )
            _, pending = await asyncio.wait(
                [
                    asyncio.create_task(sim.progress.has_reached(await_time)),
                    asyncio.create_task(sim.newer_step.wait()),
                ],
                return_when="FIRST_COMPLETED",
                timeout=world.rt_factor,
            )
            sim.newer_step.clear()
            for task in pending:
                task.cancel()
            if world.rt_factor:
                advance_progress(sim, world)
    return False


async def rt_sleep(sim: SimRunner, world: AsyncWorld) -> None:
    """
    If in real-time mode, check if to sleep and do so if necessary.
    """
    if world.rt_factor:
        rt_passed = perf_counter() - sim.rt_start
        sleep = (world.rt_factor * sim.next_steps[0].time) - rt_passed
        if sleep > 0:
            sim.tqdm.set_postfix_str("sleeping")
            await asyncio.sleep(sleep)


async def wait_for_dependencies(sim: SimRunner, lazy_stepping: bool) -> None:
    """
    Wait until all simulators that can provide input for this simulator have run for
    this step.

    Also notify any simulator that is already waiting to perform its next step.

    *world* is a mosaik :class:`~mosaik.scenario.AsyncWorld`.
    """
    futures: List[Coroutine[Any, Any, TieredTime]] = []
    next_step = sim.next_steps[0]

    for pre_sim, min_delays in sim.input_delays.items():
        # Wait for pre_sim if it hasn't progressed enough to provide
        # the input for our current step.
        for delay in min_delays.durations:
            futures.append(pre_sim.progress.has_passed(next_step, shift=delay))

    for suc_sim, adapt in sim.successors_to_wait_for.items():
        futures.append(suc_sim.progress.has_reached(next_step + adapt))
    if lazy_stepping:
        for suc_sim, adapt in sim.successors.items():
            futures.append(suc_sim.progress.has_reached(next_step + adapt))

    await asyncio.gather(*futures)


def get_input_data(world: AsyncWorld, sim: SimRunner) -> InputData:
    """
    Return a dictionary with the input data for *sim*.

    The dict will look like::

        {
            'eid': {
                'attrname': {'src_eid_0': val_0, ... 'src_eid_n': val_n},
                ...
            },
            ...
        }

    For every entity, there is an entry in the dict and each entry is itself
    a dict with attributes and a list of values. This is, because we may have
    inputs from multiple simulators (e.g., different consumers that provide
    loads for a node in a power grid) and cannot know how to aggregate that
    data (sum, max, ...?).

    *world* is a mosaik :class:`~mosaik.scenario.AsyncWorld`.
    """
    assert sim.current_step is not None
    # Input data starts with the data from set_data calls
    input_data = sim.inputs_from_set_data
    sim.inputs_from_set_data = {}
    # Merge the persistent inputs into the input data, adding keys as
    # necessary. mosaik controls three levels deep, all further levels
    # therefore should not be merged.
    merge_all(
        lambda attrs_new, attrs_old: merge_all(
            lambda data_new, data_old: merge_all(
                lambda val_new, val_old: val_new, data_new, data_old
            ),
            attrs_new,
            attrs_old,
        ),
        input_data,
        sim.persistent_inputs,
    )
    # Merge in pushed inputs from the timed input buffer
    input_data = sim.timed_input_buffer.get_input(input_data, sim.current_step.time)

    for (src_sim, delay), dataflows in sim.pulled_inputs.items():
        cache = src_sim.get_output_for(sim.current_step.time - delay.tiers[0])
        for (src_eid, src_attr), (dest_eid, dest_attr) in dataflows:
            try:
                val = cache[src_eid][src_attr]
            except KeyError:
                logger.warning(
                    f"Simulator {src_sim.sid}'s entity {src_eid} did not produce "
                    f"output on its persistent attribute {src_attr} during its last "
                    "step. However, this value is now required by simulator "
                    f"{sim.sid}. This usually results from attributes that are marked "
                    "persistent despite working like events. Supplying `None` for now. "
                    "This will be an error in future versions of mosaik."
                )
                val = None
            input_vals = input_data.setdefault(dest_eid, {}).setdefault(dest_attr, {})
            input_vals[FULL_ID % (src_sim.sid, src_eid)] = val

    # Merge the data back into the persistent inputs. Here, only keys
    # that already exist should be updated, as those are the persistent
    # attributes. (Adding others would make those persistent as well.)
    merge_existing(
        lambda attrs_old, attrs_new: merge_existing(
            lambda data_old, data_new: merge_existing(
                lambda val_old, val_new: val_new, data_old, data_new
            ),
            attrs_old,
            attrs_new,
        ),
        sim.persistent_inputs,
        input_data,
    )
    return input_data


def get_max_advance(world: AsyncWorld, sim: SimRunner, until: int) -> int:
    """
    Checks how far *sim* can safely advance its internal time during next step
    without causing a causality error.
    """
    ancs_next_steps: List[Time] = []
    for anc_sim, distances in sim.triggering_ancestors.items():
        if anc_sim.next_steps:
            for distance in distances.durations:
                ancs_next_steps.append((anc_sim.next_steps[0] + distance).time)

    own_next_step = [sim.next_steps[0].time] if sim.next_steps else []

    # The +1, -1 shenanigans exists due to how max_advance was
    # originally designed.
    return min([*ancs_next_steps, *own_next_step, until + 1]) - 1


async def step(
    world: AsyncWorld,
    sim: SimRunner,
    inputs: InputData,
    max_advance: int,
):
    """
    Advance (step) a simulator *sim* with the given *inputs*. Return an
    event that is triggered when the step was performed.

    *inputs* is a dictionary, that maps entity IDs to data dictionaries which
    map attribute names to lists of values (see :func:`get_input_data()`).

    *max_advance* is the simulation time until the simulator can safely advance
    it's internal time without causing any causality errors.
    """
    assert sim.current_step is not None
    sim.tqdm.set_postfix_str("stepping")
    sim.is_in_step = True
    next_step_time = await sim.step(sim.current_step.time, inputs, max_advance)
    sim.last_step = sim.current_step
    sim.is_in_step = False

    if next_step_time is not None:
        if not isinstance(next_step_time, int):
            raise SimulationError(
                f"the next step time returned by the step method must be of type int, "
                f'but is of type {type(next_step_time)} for simulator "{sim.sid}"'
            )
        if next_step_time <= sim.current_step.time:
            raise SimulationError(
                f"the next step time returned by step must be later than the current "
                f"step's time, but {next_step_time} <= {sim.current_step.time} "
                f'for simulator "{sim.sid}"'
            )

        if next_step_time < world.until:
            next_step_tiered_time = TieredTime(next_step_time) + sim.from_world_time
            sim.schedule_step(next_step_tiered_time)
            sim.next_self_step = next_step_tiered_time

    if sim.type == "time-based":
        assert next_step_time, "A time-based simulator must always return a next step"


def rt_check(
    rt_factor: Optional[float], rt_start: float, rt_strict: bool, sim: SimRunner
):
    """
    Check if simulation is fast enough for a given real-time factor.
    """
    if rt_factor:
        rt_passed = perf_counter() - rt_start
        delta = rt_passed - (rt_factor * sim.last_step.time)
        if delta > 0:
            if rt_strict:
                raise RuntimeError(
                    f"Simulation too slow for real-time factor {rt_factor}"
                )
            else:
                logger.warning(
                    "Simulation too slow for real-time factor {rt_factor} - {delta}s "
                    "behind time.",
                    rt_factor=rt_factor,
                    delta=delta,
                )


async def get_outputs(world: AsyncWorld, sim: SimRunner):
    """
    Wait for all required output data from a simulator *sim*.

    *world* is a mosaik :class:`~mosaik.scenario.AsyncWorld`.
    """
    assert sim.current_step is not None
    sid = sim.sid
    outattr = sim.output_request
    if outattr:
        sim.tqdm.set_postfix_str("get_data")
        data = await sim.get_data(outattr)

        output_time: int
        output_time = data.get("time", sim.last_step.time)  # type: ignore
        if output_time == sim.current_step.time:
            output_tiered_time = sim.current_step
        else:
            output_tiered_time = TieredTime(
                output_time, *([0] * (len(sim.current_step) - 1))
            )
        sim.output_time = output_tiered_time
        if sim.last_step.time > output_time:
            raise SimulationError(
                'Output time (%s) is not >= time (%s) for simulator "%s"'
                % (output_time, sim.last_step, sim.sid)
            )

        # Fill output cache. This will repeat some data that is also
        # pushed forward below, but it is faster to just save everything
        # than filter out this data here.
        if sim.outputs is not None:
            sim.outputs[output_time] = data

        # Push forward certain data
        for (src_eid, src_attr), destinations in sim.output_to_push.items():
            try:
                val = data[src_eid][src_attr]
                for dest_sim, time_shift, (dest_eid, dest_attr) in destinations:
                    dest_sim.timed_input_buffer.add(
                        output_time + time_shift.tiers[0],
                        sid,
                        src_eid,
                        dest_eid,
                        dest_attr,
                        val,
                    )
            except KeyError:
                pass
        sim.data = data


def trigger_successors(sim: SimRunner) -> None:
    """
    Notify all simulators waiting for us.
    """
    for (eid, attr), triggered in sim.triggers.items():
        if attr in sim.data.get(eid, {}):
            for dest_sim, delay in triggered:
                dest_sim.schedule_step(sim.output_time + delay)


def prune_dataflow_cache(world: AsyncWorld):
    """
    Prunes the dataflow cache.
    """
    if not world.use_cache:
        return
    min_cache_time = min(s.last_step.time for s in world.sims.values())
    for sim in world.sims.values():
        if sim.outputs:
            sim.outputs = {
                time: cache
                for time, cache in sim.outputs.items()
                if time >= min_cache_time
            }


def get_progress(sims: Dict[SimId, SimRunner], until: int) -> float:
    """
    Return the current progress of the simulation in percent.
    """
    times = [sim.progress.time.time for sim in sims.values()]
    avg_time = sum(times) / len(times)
    return avg_time * 100 / until


def get_avg_progress(sims: Dict[SimId, SimRunner], until: int) -> int:
    """Get the average progress of all simulations (in time steps)."""
    times = [min(until, sim.progress.time.time + 1) for sim in sims.values()]
    return sum(times) // len(times)


def advance_progress(sim: SimRunner, world: AsyncWorld):
    pre_sim_induced_progress: List[TieredTime] = [
        distance.earliest_sum(pre_sim.next_steps[0])
        for pre_sim, distance in sim.triggering_ancestors.items()
        if pre_sim.next_steps
    ]

    next_step_progress: List[TieredTime] = [sim.next_steps[0]] if sim.next_steps else []
    current_step_prog = [sim.current_step] if sim.current_step else []
    if world.rt_factor:
        rt_passed = perf_counter() - sim.rt_start
        rt_progress = [TieredTime(ceil(rt_passed / world.rt_factor))]
    else:
        rt_progress = []
    new_progress = min(
        [
            *pre_sim_induced_progress,
            *next_step_progress,
            *current_step_prog,
            *rt_progress,
            TieredTime(world.until) + sim.from_world_time,
        ]
    )
    sim.progress.set(new_progress)
    sim.tqdm.update(new_progress.time - sim.tqdm.n)



from __future__ import annotations

import asyncio
from typing import List, Tuple

from loguru import logger  # type: ignore  # noqa: F401

from mosaik.tiered_time import TieredDuration, TieredTime

TriggerSpec = Tuple[TieredTime, TieredDuration, bool]


class Progress:
    """A progress keeps track of a simulator's progress and provided
    coroutines for waiting for the progress to reach or pass a given
    value.

    The implemenation is generic over the type of the time values,
    provided that they are comparable.
    """

    time: TieredTime
    """The current value of the progress."""
    _futures: List[Tuple[TriggerSpec, asyncio.Future[TieredTime]]]
    """Futures representing all currently waiting has_reached and
    has_passed calls.

    The components of each tuple are the trigger time, a bool
    specifying whether the time needs to be passed (True) or just
    reached (False), and the ``Future`` to trigger upon passing/
    reaching the time.
    """

    def __init__(self, initial: TieredTime):
        self._futures = []
        self.time = initial

    def set(self, time: TieredTime):
        """Set the progress to ``value`` and trigger all waiting
        coroutines whose wait times have now been passed or reached.
        """
        assert time >= self.time, "cannot progress backwards"
        self.time = time
        # Use index-based for loop so we can call del in the loop.
        for index in reversed(range(0, len(self._futures))):
            trigger_spec, future = self._futures[index]
            triggered_time = self._triggered_time(trigger_spec)
            if triggered_time:
                if not future.cancelled():
                    future.set_result(triggered_time)
                del self._futures[index]

    def _triggered_time(self, trigger_spec: TriggerSpec) -> None | TieredTime:
        """Get the actual (destination) time at which ``trigger_spec``
        was triggered, or ``None`` if it has not been triggered yet.
        """
        target, shift, needs_to_pass = trigger_spec
        time_at_dest = self.time + shift
        if needs_to_pass and time_at_dest > target:
            return time_at_dest
        if not needs_to_pass and time_at_dest >= target:
            return time_at_dest
        return None

    async def _add_trigger(
        self, target: TieredTime, shift: TieredDuration | None, needs_to_pass: bool
    ) -> TieredTime:
        """Add a trigger to this progress. This gets called by
        ``has_reached`` and ``has_passed`` with ``needs_to_pass`` set to
        ``False`` or ``True``, respectively.
        """
        if shift is None:
            shift = TieredDuration(*((0,) * len(self.time)))
        trigger_spec = (target, shift, needs_to_pass)
        triggered_time = self._triggered_time(trigger_spec)
        if triggered_time:
            return triggered_time
        future: asyncio.Future[TieredTime] = asyncio.Future()
        self._futures.append((trigger_spec, future))
        return await future

    async def has_reached(
        self,
        target: TieredTime,
        shift: TieredDuration | None = None,
    ) -> TieredTime:
        """Wait until this ``Progress`` has reached (or passed) the
        given time. Returns immediately if this ``Progress`` has
        already reached (or passed) the given time.

        :param time: the time that needs to be reached (or passed)
        :returns: the actual value of the progress when the given
            time has been reached or passed
        """
        return await self._add_trigger(target, shift, False)

    async def has_passed(
        self,
        target: TieredTime,
        shift: TieredDuration | None = None,
    ) -> TieredTime:
        """Wait until this ``Progress`` has passed the given time.
        Returns immediately if this ``Progress`` has already passed
        the given time.

        :param time: the time that needs to be passed
        :return: the actual value of the progress when the given
            time has been passed
        """
        return await self._add_trigger(target, shift, True)

    def __repr__(self) -> str:
        return f"<Progress at {self.time!r} with {len(self._futures)} waiting>"



from typing import Any, Callable, Dict, List

import mosaik_api_v3
from mosaik_api_v3.types import (
    CreateResult,
    InputData,
    Meta,
    ModelName,
    OutputData,
    OutputRequest,
    SimId,
    Time,
)
from typing_extensions import override

FUNCTION_KEY = "Function"
CONSTANT_KEY = "Constant"

META: Meta = {
    "api_version": "3.0",
    "type": "time-based",
    "models": {
        FUNCTION_KEY: {
            "public": True,
            "params": ["function"],
            "attrs": ["value"],
        },
        CONSTANT_KEY: {
            "public": True,
            "params": ["constant"],
            "attrs": ["value"],
        },
    },
    "extra_methods": [],
}


class InputSimulator(mosaik_api_v3.Simulator):
    """
    This simulator gives a steady input to a connected simulator.
    This input can either be an constant value or given by a custom
    function based on the current time.

    When starting the simulator, a custom step size may be provided
    using the `step_size` parameter. The default is 1.

    When creating a `Constant` entity, the constant must be passed as
    the parameter `constant`.

    When creating a `Function` entity, a function should be passed as
    the `function` parameter. This function should take the current
    mosaik time and return the desired value. This type of entity only
    works when the simulator is started using the `"python"` method.

    In either case, the entity will produce its output on the *value*
    attribute.
    """

    step_size: int
    functions: Dict[str, Callable[[Time], Any]]
    constants: Dict[str, Any]

    def __init__(self):
        super().__init__(META)
        self.functions = {}
        self.constants = {}
        self.step_size = 1
        self.time = 0

    @override
    def init(self, sid: SimId, time_resolution: float = 1, step_size: int = 1) -> Meta:
        self.step_size = step_size
        return self.meta

    @override
    def create(
        self, num: int, model: ModelName, **model_params: Any
    ) -> List[CreateResult]:
        new_entities: List[CreateResult] = []
        if model == FUNCTION_KEY:
            for i in range(len(self.functions), len(self.functions) + num):
                new_entities.append(
                    self.create_function_entity(i, model, **model_params)
                )
        elif model == CONSTANT_KEY:
            for i in range(len(self.constants), len(self.constants) + num):
                new_entities.append(
                    self.create_constant_entity(i, model, **model_params)
                )
        return new_entities

    def create_function_entity(
        self, id: int, model: str, function: Callable[[Time], Any]
    ) -> CreateResult:
        eid = f"{FUNCTION_KEY}-{id}"
        self.functions[eid] = function
        return {
            "eid": eid,
            "type": model,
        }

    def create_constant_entity(
        self, id: int, model: str, constant: Any
    ) -> CreateResult:
        eid = f"{CONSTANT_KEY}-{id}"
        self.constants[eid] = constant
        return {"eid": eid, "type": model}

    @override
    def step(self, time: Time, inputs: InputData, max_advance: Time) -> Time:
        assert inputs == {}
        self.time = time
        return time + self.step_size

    @override
    def get_data(self, outputs: OutputRequest) -> OutputData:
        data: OutputData = {}
        for eid in outputs:
            if CONSTANT_KEY in eid:
                data[eid] = {"value": self.constants[eid]}
            else:
                data[eid] = {"value": self.functions[eid](self.time)}
        return data



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from copy import deepcopy
from inspect import isgeneratorfunction
from subprocess import Popen
from typing import Any, Dict, Iterator, List, Tuple

from loguru import logger
from mosaik_api_v3 import MosaikProxy, Simulator, check_api_compliance
from mosaik_api_v3.connection import Channel, EndOfRequests
from mosaik_api_v3.types import Meta, SimId

from mosaik.exceptions import ScenarioError


class Proxy(ABC):
    """A proxy for a mosaik simulator from the view of a mosaik
    scenario.

    Generally, this will be a ``BaseProxy`` subclass wrapped in the
    appropriate ``Adapter`` subclasses to bring the interface of the
    connected simulator in line with the most up-to-date API version.
    """

    @abstractmethod
    async def send(self, request: Any) -> Any:
        """Send a request to the connected simulator.

        :param request: Generally, this will be a three-tuple consisting
        of a function name, a list of positional arguments and a dict
        of named arguments.
        :return: The return value from the remote simulator (depends on
        the specified function).
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def meta(self) -> Meta:
        """The meta of the connected simulator, as adapted by the
        adapters.
        """
        raise NotImplementedError()

    @abstractmethod
    async def stop(self) -> None:
        """Stop the connected simulator. This is not handled via
        ``send`` as there are extra steps to be taken to close the
        connection cleanly.
        """
        raise NotImplementedError()


class BaseProxy(Proxy):
    """A base ``Proxy`` for a connected simulator that simply sends all
    requests along unchanged. This will usually be wrapped in one or
    more ``Adapter``s to allow treating the simulator as up-to-date from
    other parts of mosaik.
    """

    @abstractmethod
    async def init(
        self, sid: SimId, *, time_resolution: float, **sim_params: Any
    ) -> List[int]:
        """Initialize the simulator by sending the ``init`` call. The
        ``meta`` returned by the simulator will be saved to be retrieved
        using the ``meta`` property.

        :param sid: The ``SimId`` that mosaik assigns to this simulator
        instance
        :param time_resolution: The time resolution of the simulation,
        i.e. how many seconds correspond to one mosaik time step.
        :param sim_params: The params sent to the simulator for
        initialization.
        """
        raise NotImplementedError()


class LocalProxy(BaseProxy):
    """
    Proxy for a local simulator. This mainly wraps each mosaik method in
    a coroutine.
    """

    sim: Simulator
    """The underlying ``mosaik_api.Simulator."""

    def __init__(self, sim: Simulator, mosaik_remote: MosaikProxy):
        super().__init__()
        self.sim = sim
        sim.mosaik = mosaik_remote

    async def init(self, sid: SimId, **kwargs: Any) -> List[int]:
        # This in an ugly place for these checks. However, we cannot
        # put them in mosaik.adapters because we need to determine
        # API compliance before sending the init method and thus before
        # receiving the version number to build the adapter.
        if check_api_compliance(self.sim):
            forced_old_api = False
        else:
            forced_old_api = True
            del kwargs["time_resolution"]

        meta = await self.send(("init", (sid,), kwargs))
        self._meta = deepcopy(meta)
        version = extract_version(meta)
        if forced_old_api and version >= [3]:
            raise ScenarioError(
                "The underlying simulator is not compliant with the high-level API "
                "version 3 (or higher) (because its init method is missing the "
                "time_resolution keyword parameter or its step method is missing the "
                "max_advance parameter), but it claims to be of version "
                f"{'.'.join(map(str, version))} in its meta's api_version field."
            )
        return version

    @property
    def meta(self):
        return self._meta

    async def send(self, request: Tuple[str, Tuple[Any, ...], Dict[str, Any]]):
        func_name, args, kwargs = request
        func = getattr(self.sim, func_name)
        # A simulator that makes requests back to mosaik (like set_data or set_event)
        # will have generator functions instead of normal functions as its init, create,
        # step and/or get_data. It will yield coroutines that produce the required
        # information, which we have to await. (This is due to simpy, which used
        # generator functions for its asynchronicity; we didn't want to break the API.)
        # TODO: Maybe check this during __init__ and create the right methods instead of
        # checking for isgeneratorfunction on each call?
        if isgeneratorfunction(func):
            gen = func(*args, **kwargs)
            try:
                incoming_request = next(gen)
                while True:
                    incoming_request = gen.send(await incoming_request)
            except StopIteration as stop:
                return stop.value
        else:
            return func(*args, **kwargs)

    async def stop(self):
        self.sim.finalize()


class RemoteProxy(BaseProxy):
    _channel: Channel
    _reader_task: asyncio.Task[None]
    _outgoing_msg_counter: Iterator[int]
    _mosaik_remote: MosaikProxy
    _process: Tuple[Popen[str], bool] | None
    """The process for this RemoteProxy (or None, if the connection
    was established using connect). The second component of the tuple is
    True if the process should be automatically terminated at the end of
    the simulation.
    """

    def __init__(
        self,
        channel: Channel,
        mosaik_remote: MosaikProxy,
        *,
        process: Tuple[Popen[str], bool] | None = None,
    ):
        super().__init__()
        self._channel = channel
        self._mosaik_remote = mosaik_remote
        self._reader_task = asyncio.create_task(
            self._handle_remote_requests(), name="handle remote requests for ???"
        )
        self._process = process

    async def _handle_remote_requests(self) -> None:
        try:
            while True:
                request = await self._channel.next_request()
                func_name, args, kwargs = request.content
                func = getattr(self._mosaik_remote, func_name)
                try:
                    result = await func(*args, **kwargs)
                    await request.set_result(result)
                except Exception as e:
                    await request.set_exception(e)
        except EndOfRequests:
            pass
        except RuntimeError as e:
            if e.args[0] != "Event loop is closed":
                logger.exception(
                    "Something went wrong in _handle_remote_requests, "
                    f"exception type {type(e)}"
                )
                await self.stop()
        except Exception as e:
            logger.exception(
                "Something went wrong in _handle_remote_requests, "
                f"exception type {type(e)}"
            )
            await self.stop()

    async def init(self, sid: SimId, **kwargs: Any) -> List[int]:
        self._meta = await self.send(["init", (sid,), kwargs])
        return extract_version(self._meta)

    @property
    def meta(self) -> Meta:
        return self._meta

    async def send(self, request: Any) -> Any:
        return await self._channel.send(request)

    async def stop(self) -> None:
        try:
            await asyncio.wait_for(self._channel.send(["stop", [], {}]), 0.1)
        except (asyncio.TimeoutError, asyncio.IncompleteReadError):
            pass
        await self._channel.close()
        await self._reader_task
        if self._process and self._process[1]:
            self._process[0].terminate()


def extract_version(meta: Meta) -> List[int]:
    if "api_version" not in meta:
        return [1]
    else:
        return list(map(int, meta["api_version"].split(".")))



"""
This module allows you to activate some debugging functionality that makes
mosaik collect more data when the simulation is being executed.
"""

from __future__ import annotations

from copy import deepcopy
from time import perf_counter
from typing import Dict, List, Optional, Tuple

import networkx as nx
from loguru import logger  # noqa: F401  # type: ignore
from mosaik_api_v3 import InputData, SimId

from mosaik import AsyncWorld, scheduler
from mosaik.scenario import World
from mosaik.simmanager import SimRunner
from mosaik.tiered_time import TieredDuration, TieredTime

_originals = {
    "step": scheduler.step,
}


def enable():
    """
    Wrap :func:`~mosaik.scheduler.step()` to collect more data about the
    scheduler execution.
    """

    async def wrapped_step(
        world: AsyncWorld, sim: SimRunner, inputs: InputData, max_advance: int
    ):
        pre_step(world, sim, inputs)
        ret = await _originals["step"](world, sim, inputs, max_advance)
        post_step(world, sim)
        return ret

    scheduler.step = wrapped_step
    World.assert_graph = assert_graph
    World.assert_inputs = assert_inputs


def disable():
    """
    Restore all wrapped functions to their original.
    """
    for k, v in _originals.items():
        setattr(scheduler, k, v)


def parse_node(node_str: str) -> Tuple[SimId, TieredTime]:
    # networkx will call the parser on already-parsed nodes occasionally
    # So we make sure that we only try to parse strings.
    if isinstance(node_str, str):
        sid, time = node_str.rsplit("~", 1)
        return (sid, TieredTime(*tuple(map(int, time.split(":")))))
    return node_str


def parse_execution_graph(graph_string: str) -> nx.DiGraph[Tuple[SimId, TieredTime]]:
    return nx.parse_edgelist(
        graph_string.split("\n"),
        create_using=nx.DiGraph(),
        nodetype=parse_node,
        data=(),
    )


def pre_step(world: AsyncWorld, sim: SimRunner, inputs: InputData):
    """
    Add a node for the current step and edges from all dependencies to the
    :attr:`mosaik.scenario.World.execution_graph`.

    Also perform some checks and annotate the graph with the dataflows.
    """
    eg = world.execution_graph
    sims = world.sims

    sid = sim.sid
    assert sim.current_step is not None
    next_step = sim.current_step
    node_id = (sid, next_step)

    sim.last_node = node_id

    eg.add_node(node_id, t=perf_counter(), inputs=deepcopy(inputs))

    input_pres = {
        kk.split(".")[0]
        for ii in inputs.values()
        for jj in ii.values()
        for kk in jj.keys()
    }
    for pre_sim, min_delays in sim.input_delays.items():
        if pre_sim.sid in input_pres or sim in pre_sim.successors_to_wait_for:
            pre_node: Optional[Tuple[str, TieredTime]] = None
            pre_time = TieredTime(-1, *([0] * (len(pre_sim.progress.time) - 1)))
            # We check for all nodes if it is from the predecessor and it its
            # step time is before the current step of sim. There might be cases
            # where this simple procedure is wrong, e.g. when the pred has
            # stepped but didn't provide the connected output.
            for inode in eg.nodes:
                node_sid, itime = inode
                if (
                    node_sid == pre_sim.sid
                    and next_step >= min_delays.earliest_sum(itime)
                    and itime >= pre_time
                ):
                    pre_node = inode
                    pre_time = itime
            if pre_node is not None:
                eg.add_edge(pre_node, node_id)
                assert eg.nodes[pre_node]["t"] <= eg.nodes[node_id]["t"]

    for suc_sim in sim.successors_to_wait_for:
        suc = suc_sim.sid
        if sim.last_step >= TieredTime(0):
            suc_node = (suc, sims[suc].last_step)
            eg.add_edge(suc_node, node_id)
            assert sims[suc].progress.time + TieredDuration(1) >= next_step


def post_step(world: AsyncWorld, sim: SimRunner):
    """
    Record time after a step and add self-step edge.
    """
    eg = world.execution_graph
    last_node = sim.last_node
    eg.nodes[last_node]["t_end"] = perf_counter()
    next_self_step = sim.next_self_step
    if (
        next_self_step is not None
        and next_self_step < TieredTime(world.until) + sim.from_world_time
    ):
        node_id = (sim.sid, next_self_step)
        eg.add_edge(sim.last_node, node_id)
        sim.next_self_step = None


def assert_graph(world: AsyncWorld, expected_str: str, extra_nodes: List[str] = []):  # noqa: C901
    actual_graph = world.execution_graph
    expected_graph = parse_execution_graph(expected_str)
    for node in extra_nodes:
        expected_graph.add_node(parse_node(node))

    errors: List[str] = []
    expected_nodes = set(expected_graph.nodes)
    actual_nodes = set(actual_graph.nodes)
    missing_nodes = expected_nodes - actual_nodes

    def format_node(node: Tuple[str, TieredTime]) -> str:
        return f"{node[0]} @ {node[1]}"

    if missing_nodes:
        errors.append("The following expected simulator invocations did not happen:")
        for node in sorted(missing_nodes):
            errors.append(f"- {format_node(node)}")
        errors.append("")

    unexpected_nodes = actual_nodes - expected_nodes
    if unexpected_nodes:
        errors.append("The following simulator invocations were not expected:")
        for node in sorted(unexpected_nodes):
            sources = actual_graph.predecessors(node)
            if sources:
                sources_str = f"caused by: {', '.join(map(format_node, sources))}"
            else:
                sources_str = "not caused by other simulators"
            errors.append(f"- {format_node(node)} ({sources_str})")
        errors.append("")

    predecessor_errors: List[str] = []
    for node in sorted(actual_nodes & expected_nodes):
        actual_pres = set(actual_graph.predecessors(node))
        expected_pres = set(expected_graph.predecessors(node))
        if actual_pres != expected_pres:
            predecessor_errors.append(
                f"- {format_node(node)} ("
                f"extraneous {', '.join(map(format_node, sorted(actual_pres - expected_pres)))}; "
                f"missing {', '.join(map(format_node, sorted(expected_pres - actual_pres)))})"
            )
    if predecessor_errors:
        errors.append("The following simulator invocations had incorrect sources:")
        errors.extend(predecessor_errors)
        errors.append("")

    if errors:
        raise AssertionError(
            "The following problems were detected in the execution graph:\n\n"
            + "\n".join(errors)
        )

    assert actual_graph.adj == expected_graph.adj


def assert_inputs(world: AsyncWorld, expected_inputs: Dict[str, InputData]):
    eg = world.execution_graph
    for node_str, expected_data in expected_inputs.items():
        node = parse_node(node_str)
        assert expected_data == eg.nodes[node]["inputs"]
        del eg.nodes[node]["inputs"]
    for node in eg.nodes(data="inputs"):
        assert not node[1]  # Make sure that there are not unchecked inputs



# pyright: reportUnknownMemberType=false
"""
This module contains some utility functions and classes.

"""

from __future__ import annotations

import datetime
import random
from typing import (
    TYPE_CHECKING,
    Collection,
    Dict,
    Iterable,
    List,
    MutableSequence,
    Optional,
    Set,
    Tuple,
)

import networkx as nx
from mosaik_api_v3 import Attr, SimId
from typing_extensions import Literal

from mosaik.async_scenario import AsyncWorld
from mosaik.scenario import Entity, World
from mosaik.tiered_time import TieredTime

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure

STANDARD_DPI = 600
STANDARD_FORMAT = "png"
STANDARD_FOLDER = "figures"


def connect_many_to_one(
    world: World | AsyncWorld,
    src_set: Iterable[Entity],
    dest: Entity,
    *attrs: Attr | Tuple[Attr, Attr],
    async_requests: bool = False,
):
    """:meth:`~mosaik.scenario.World.connect` each entity in *src_set*
    to *dest*.

    See the :meth:`~mosaik.scenario.World.connect` for more details.
    """
    for src in src_set:
        world.connect(src, dest, *attrs, async_requests=async_requests)


def connect_randomly(
    world: World | AsyncWorld,
    src_set: MutableSequence[Entity],
    dest_set: MutableSequence[Entity],
    *attrs: Attr | Tuple[Attr, Attr],
    evenly: bool = True,
    max_connects: int = float("inf"),  # type: ignore
):
    """
    Randomly :meth:`~mosaik.scenario.World.connect` the entities from
    *src_set* to the entities from *dest_set* and return a subset of *dest_set*
    containing all entities with a connection.

    *world* is an instance of the :class:`~mosaik.scenario.World` to which the
    entities belong.

    *src_set* and *dest_set* are iterables containing
    :class:`~mosaik.scenario.Entity` instances. *src_set* may be empty,
    *dest_set* must not be empty. Each entity of *src_set* will be connected to
    an entity of *dest_set*, but not every entity of *dest_set* will
    necessarily have a connection (e.g., if you connect a set of three entities
    to a set of four entities). A set of all entities from *dest_set*, to which
    at least one entity from *src_set* was connected, will be returned.

    *attrs* is a list of attribute names of pairs as in
    :meth:`~mosaik.scenario.World.connect()`.

    If the flag *evenly* is set to ``True``, entities connections will be
    distributed as evenly as possible. That means if you connect a set of three
    entities to a set of three entities, there will be three 1:1 connections;
    if you connect four entities to three entities, there will be one 2:1 and
    two 1:1 connections. If *evenly* is set to ``False``, connections will be
    truly random. That means if you connect three entities to three entities,
    you may either have three 1:1 connections, one 2:1 and two 1:1 connections
    or just one 3:1 connection.

    *max_connects* lets you set the maximum number of connections that an
    entity of *dest_set* may receive. This argument is only taken into account
    if *evenly* is set to ``False``.
    """
    dest_set = list(dest_set)
    assert dest_set

    if evenly:
        connected = _connect_evenly(world, src_set, dest_set, *attrs)
    else:
        connected = _connect_randomly(
            world, src_set, dest_set, *attrs, max_connects=max_connects
        )

    return connected


def _connect_evenly(
    world: World | AsyncWorld,
    src_set: MutableSequence[Entity],
    dest_set: MutableSequence[Entity],
    *attrs: Attr | Tuple[Attr, Attr],
) -> Set[Entity]:
    connect = world.connect
    connected: Set[Entity] = set()

    src_size, dest_size = len(src_set), len(dest_set)
    pos = 0
    while pos < src_size:
        random.shuffle(dest_set)
        for src, dest in zip(src_set[pos:], dest_set):
            connect(src, dest, *attrs)
            connected.add(dest)
        pos += dest_size

    return connected


def _connect_randomly(
    world: World | AsyncWorld,
    src_set: MutableSequence[Entity],
    dest_set: MutableSequence[Entity],
    *attrs: Attr | Tuple[Attr, Attr],
    max_connects: int = float("inf"),  # type: ignore
) -> Set[Entity]:
    connect = world.connect
    connected: Set[Entity] = set()

    assert len(src_set) <= (len(dest_set) * max_connects)
    max_i = len(dest_set) - 1
    randint = random.randint
    connects: Dict[Entity, int] = {}
    for src in src_set:
        i = randint(0, max_i)
        dest = dest_set[i]
        connect(src, dest, *attrs)
        connected.add(dest)
        connects[dest] = connects.get(dest, 0) + 1
        if connects[dest] >= max_connects:
            dest_set.remove(dest)
            max_i -= 1
            assert max_i >= 0

    return connected


def plot_execution_time(
    world: World,
    folder: str = STANDARD_FOLDER,
    hdf5path: str | None = None,
    dpi: int = STANDARD_DPI,
    format: Literal["png", "pdf", "svg"] = STANDARD_FORMAT,
    show_plot: bool = True,
    slice: Tuple[int, int] | None = None,
):
    """Creates an image visualizing the execution time of the different
    simulators of a mosaik scenario.

    :param world: mosaik world object
    :param folder: folder to store the image (only if no hdf5path is
        provided)
    :param hdf5path: Path to HDF5 file, which will be used as path for
        the created image
    :param dpi: DPI for created images
    :param format: format for created image
    :param show_plot: whether to open a window to show the plot
    :param slice: reduce the timeframe that you show in the plot. Usage
        as in Python list slicing, i.e., negative values are possible to
        start from the end of the list. Jumps are not possible.
        ``slice`` needs to be a two-element integer list, e.g.
        ``(0, 5)``.

    :return: ``None`` but image file will be written to file system
    """
    import matplotlib.pyplot as plt

    steps = {}
    all_nodes = list(world.execution_graph.nodes(data=True))

    # Slice the data if the slice reduces the timesteps to be shown
    if slice is not None:
        slices_steps = range(world.until)[slice[0] : slice[1]]
        all_nodes_sliced = []
        for node in all_nodes:
            if node[0][1].time in slices_steps:
                all_nodes_sliced.append(node)
        all_nodes = all_nodes_sliced

    t_min = min([node[1]["t"] for node in all_nodes])
    t_max = max([node[1]["t_end"] for node in all_nodes])

    for isid in world.sims.keys():
        steps[isid] = []
        for node in all_nodes:
            if node[0][0] == isid:
                steps[isid].append(
                    (node[1]["t"] - t_min, (node[1]["t_end"] - node[1]["t"]))
                )

    fig, ax = plt.subplots()
    for i, isid in enumerate(world.sims.keys()):
        ax.broken_barh(steps[isid], ((i + 1) * 10, 9), facecolors="tab:blue")
    ax.set_xlim(0, t_max - t_min)
    ax.set_ylim(5, len(world.sims.keys()) * 10 + 15)
    ax.set_yticks(list(range(15, len(world.sims.keys()) * 10 + 10, 10)))
    ax.set_yticklabels(list(world.sims.keys()))
    ax.set_xlabel("Simulation time [s]")
    ax.grid(True)
    if hdf5path:
        filename = hdf5path.replace(".hdf5", "_graph." + format)
    else:
        filename: str = get_filename(folder, "executiontime", format)

    fig.savefig(
        filename,
        format=format,
        dpi=dpi,
        facecolor="white",
        transparent=True,
        bbox_inches="tight",
    )
    if show_plot is True:
        plt.show()


def plot_dataflow_graph(
    world: World,
    folder: str = STANDARD_FOLDER,
    hdf5path: Optional[str] = None,
    dpi: int = STANDARD_DPI,
    format: Literal["png", "pdf", "svg"] = STANDARD_FORMAT,
    show_plot: bool = True,
):
    """Creates an image visualizing the data flow graph of a mosaik
    scenario. Using the spring layout from Matplotlib (Fruchterman-
    Reingold force-directed algorithm) to position the nodes.

    :param world: mosaik world object
    :param folder: folder to store the image (only if no hdf5path is
        provided)
    :param hdf5path: Path to HDF5 file, which will be used as path for
        the created image
    :param dpi: DPI for created images
    :param format: format for created image
    :param show_plot: whether open a window to show the plot
    :return: ``None`` but image file will be written to file
        system
    """
    import matplotlib.pyplot as plt
    from matplotlib.patches import ConnectionPatch

    # Recreate the df_graph for plotting. There might be additional
    # useful information to be extracted from the SimRunners.
    df_graph: nx.DiGraph[str] = nx.DiGraph()
    for sim in world.sims.values():
        df_graph.add_node(sim.sid)
        for pred, delay in sim.input_delays.items():
            df_graph.add_edge(
                pred.sid,
                sim.sid,
                time_shifted=delay.tiers[0] > 0,
                weak=any(t > 0 for t in delay.tiers[1:]),
            )
    positions = nx.spring_layout(df_graph)

    fig, ax = plt.subplots()
    for node in df_graph.nodes:
        # Draw a dot for the simulator
        ax.plot(positions[node][0], positions[node][1], "o")
        # Put the name of the simulator on the dot. If we put an
        # absolute distance, we depend on the scaling, which can effect
        # seemingly random distances from the dot
        text_x = positions[node][0]
        text_y = positions[node][1]
        label = ax.annotate(node, positions[node], xytext=(text_x, text_y), size=4)
        label.set_alpha(0.6)

    for edge in list(df_graph.edges()):
        edge_infos = df_graph.adj[edge[0]][edge[1]]
        annotation = ""
        color = "grey"
        linestyle = "solid"
        if edge_infos["time_shifted"]:
            color = "tab:red"
            annotation = "time_shifted"

        if edge_infos["weak"]:
            annotation += " weak"
            linestyle = "dotted"

        x_pos0 = positions[edge[0]][0]
        x_pos1 = positions[edge[1]][0]
        y_pos0 = positions[edge[0]][1]
        y_pos1 = positions[edge[1]][1]

        con = ConnectionPatch(
            (x_pos0, y_pos0),
            (x_pos1, y_pos1),
            "data",
            "data",
            arrowstyle="->",
            linestyle=linestyle,
            connectionstyle="arc3,rad=0.1",
            shrinkA=5,
            shrinkB=5,
            mutation_scale=20,
            fc="w",
            color=color,
            alpha=0.6,
        )
        ax.add_artist(con)

        # Attention: This is not the actual mid-point in the line
        # I suspect its more like a control point in a bezier interpolation
        # When the line is more curved, the middle point here is further away from the actual line
        # One could suspect that the mid-point is actually the middle point in this array,
        # but the array starts with the stating point, then has the curve-control point in the middle
        # and then has the points that draw the arrow
        # Why not calculating the middle point on the straight line? Because then by a 50/50 chance
        # when you have a curved arrow back and forth between two points, you can have the annotation
        # above the wrong arrow.
        midpoint: Tuple[float, float] = con.get_path().vertices[1]  # type: ignore  # close enough

        ax.annotate(
            annotation,
            (midpoint[0], midpoint[1]),
            xytext=(0, 0),
            textcoords="offset points",
            color=color,
            fontsize=5,
        )

    plt.axis("off")

    if show_plot is True:
        plt.show()

    if hdf5path:
        filename: str = hdf5path.replace(".hdf5", "graph_df." + format)
    else:
        filename: str = get_filename(folder, "dataflowGraph_2", format)

    fig.savefig(
        filename,
        format=format,
        dpi=dpi,
        facecolor="white",
        transparent=True,
        bbox_inches="tight",
    )


def plot_execution_graph(  # noqa: C901
    world: World,
    title: str = "",
    folder: str = STANDARD_FOLDER,
    hdf5path: str | None = None,
    dpi: int = STANDARD_DPI,
    format: Literal["png", "pdf", "svg"] = STANDARD_FORMAT,
    show_plot: bool = True,
    save_plot: bool = True,
    slice: Tuple[int, int] | None = None,
):
    """Creates an image visualizing the execution graph of a mosaik
    scenario.

    :param world: mosaik world object
    :param title: the title of the graph
    :param folder: folder to store the image (only if no hdf5path is
        provided)
    :param hdf5path: Path to HDF5 file, which will be used as path for
        the created image
    :param dpi: DPI for created images
    :param format: format for created image
    :param show_plot: whether to open a window to show the plot
    :param slice: reduce the timeframe that you show in the plot.
        Usage as in Python list slicing, i.e., negative values are
        possible to start from the end of the list. Jumps are not
        possible. ``slice`` needs to be a two-element integer tuple,
        e.g. ``(0, 5)``.

    :return: ``None`` but image file will be written to file system
    """
    import matplotlib.pyplot as plt
    from matplotlib import rcParams
    from matplotlib.ticker import MaxNLocator

    all_nodes = list(world.execution_graph.nodes(data=True))

    rcParams.update({"figure.autolayout": True})

    steps_st: Dict[SimId, List[float]] = {}
    for sim_name in world.sims.keys():
        steps_st[sim_name] = []

    for node in all_nodes:
        sim_name, tiered_time = node[0]
        steps_st[sim_name].append(_tiered_time_pos(tiered_time))

    fig, ax = plt.subplots()
    if title:
        fig.suptitle(title)

    # Draw the time steps from the simulators
    number_of_steps = 0
    colormap = ["black" for _ in world.sims]
    for i, sim_name in enumerate(world.sims):
        # We need the number of steps in the simulation for correct plotting with slices
        if number_of_steps < len(steps_st[sim_name]):
            number_of_steps = len(steps_st[sim_name])

        if slice is not None:
            dot = ax.plot(
                steps_st[sim_name][slice[0] : slice[1]],
                [i] * len(steps_st[sim_name][slice[0] : slice[1]]),
                "o",
            )
        else:
            dot = ax.plot(steps_st[sim_name], [i] * len(steps_st[sim_name]), "o")
        # Store the color that is used for the dots in this line (for this simulator)
        colormap[i] = dot[0].get_color()

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set_yticks(list(range(len(world.sims.keys()))))
    ax.set_yticklabels(list(world.sims.keys()))

    all_edges = list(world.execution_graph.edges())
    y_pos: Dict[SimId, int] = {}
    for sim_count, sim_name in enumerate(world.sims.keys()):
        y_pos[sim_name] = sim_count

    # The slice values can be negative, so we want to have the correct time steps
    labels = None
    if slice is not None:
        labels = range(world.until)[slice[0] : slice[1]]

    for edge in all_edges:
        isid_0, t0 = edge[0]
        isid_1, t1 = edge[1]

        if arrow_is_not_in_slice(labels, t0.time, t1.time):
            continue

        x_pos0 = _tiered_time_pos(t0)
        x_pos1 = _tiered_time_pos(t1)
        y_pos0 = y_pos[isid_0]
        y_pos1 = y_pos[isid_1]

        ax.annotate(
            "",
            (x_pos1, y_pos1),
            xytext=(x_pos0, y_pos0),
            arrowprops={
                "color": colormap[y_pos0],
                "arrowstyle": "->",
                "connectionstyle": "arc3,rad=0.05",
                "alpha": 0.6,
            },
        )

    if show_plot is True:
        plt.show()

    if hdf5path:
        filename: str = hdf5path.replace(".hdf5", "graph_execution." + format)
    else:
        filename: str = get_filename(folder, "executionGraph", format)

    if save_plot:
        fig.savefig(
            filename,
            format=format,
            dpi=dpi,
            facecolor="white",
            transparent=True,
            bbox_inches="tight",
        )


def arrow_is_not_in_slice(
    labels: Collection[int] | None,
    t0: int,
    t1: int,
):
    return labels is not None and (t0 not in labels or t1 not in labels)


def plot_execution_time_per_simulator(
    world: World,
    folder: str = STANDARD_FOLDER,
    hdf5path: str | None = None,
    dpi: int = STANDARD_DPI,
    format: Literal["png", "pdf", "svg"] = STANDARD_FORMAT,
    show_plot: bool = True,
    plot_per_simulator: bool = False,
    slice: Tuple[int, int] | None = None,
):
    """Creates images visualizing the execution time of each of the
    different simulators of a mosaik scenario.

    :param world: mosaik world object
    :param folder: folder to store the image (only if no hdf5path is
        provided)
    :param hdf5path: Path to HDF5 file, which will be used as path for
        the created image
    :param dpi: DPI for created images
    :param format: format for created image
    :param show_plot: whether to open a window to show the plot
    :param plot_per_simulator: whether to create a separated plot per
        simulator. This is especially useful if the step sizes of the
        simulators are very different.
    :param slice: reduce the timeframe that you show in the plot. Usage
        as in Python list slicing, i.e., negative values are possible to
        start from the end of the list. Jumps are not possible.
        ``slice`` needs to be a two-element integer tuple, e.g.
        ``(0, 5)``.
    :return: ``None`` but image file will be written to file system
    """
    execution_graph = world.execution_graph
    results: Dict[SimId, List[float]] = {}
    for node in execution_graph.nodes:
        execution_time = (
            execution_graph.nodes[node]["t_end"] - execution_graph.nodes[node]["t"]
        )
        sim_id = node[0]
        results.setdefault(sim_id, []).append(execution_time)

    if plot_per_simulator is False:
        fig, sub_figure = init_execution_time_per_simulator_plot()
        for key in results.keys():
            plot_results = get_execution_time_per_simulator_plot_data(
                slice, results, sub_figure, key
            )
            sub_figure.plot(plot_results, label=key)
        finish_execution_time_per_simulator_plot(
            folder, hdf5path, dpi, format, show_plot, fig
        )
    else:
        for key in results.keys():
            fig, sub_figure = init_execution_time_per_simulator_plot()
            plot_results = get_execution_time_per_simulator_plot_data(
                slice, results, sub_figure, key
            )
            sub_figure.plot(plot_results, label=key)
            finish_execution_time_per_simulator_plot(
                folder, hdf5path, dpi, format, show_plot, fig, "_" + key
            )


def get_execution_time_per_simulator_plot_data(
    slice: Tuple[int, int] | None,
    results: Dict[SimId, List[float]],
    sub_figure: Axes,
    key: SimId,
) -> List[float]:
    if slice is not None:
        plot_results = results[key][slice[0] : slice[1]]
        # The slice values can be negative, so we want to have the correct time steps
        labels = range(len(results[key]))[slice[0] : slice[1]]
        sub_figure.set_xticks(range(0, len(labels)), map(str, labels))
    else:
        plot_results = results[key]
    return plot_results


def finish_execution_time_per_simulator_plot(
    folder: str,
    hdf5path: str | None,
    dpi: int,
    format: Literal["png", "svg", "pdf"],
    show_plot: bool,
    fig: Figure,
    simulator_name: str = "",
):
    import matplotlib.pyplot as plt

    fig.legend()
    if hdf5path:
        filename: str = hdf5path.replace(".hdf5", "_" + "all" + ".png")
    else:
        filename: str = get_filename(
            folder, "execution_time_simulator" + simulator_name, format
        )

    fig.savefig(
        filename,
        format=format,
        dpi=dpi,
        facecolor="white",
        transparent=True,
        bbox_inches="tight",
    )

    if show_plot is True:
        plt.show()

    plt.close()


def init_execution_time_per_simulator_plot() -> Tuple[Figure, Axes]:
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MaxNLocator

    fig: Figure = plt.figure()
    sub_figure: Axes = fig.add_subplot()
    sub_figure.set_title("Execution time")
    sub_figure.set_ylabel("Execution time [s]")
    sub_figure.set_xlabel("Simulation time [steps of the simulator]")
    sub_figure.get_xaxis().set_major_locator(MaxNLocator(integer=True))
    return fig, sub_figure


def get_filename(dir: str, type: str, file_format: str) -> str:
    return (
        dir
        + "/"
        + str(datetime.datetime.now())
        .replace(" ", "")
        .replace(":", "")
        .replace(".", "")
        + "_"
        + type
        + "."
        + file_format
    )


def _tiered_time_pos(time: TieredTime, base: float = 0.1) -> float:
    result = 0.0
    factor = 1.0
    for tier in time.tiers:
        result += factor * tier
        factor *= base
    return result



# This module provides a way of introducing new versions of the mosaik
# API without having to check the version of the API everywhere in the
# code. The basic idea is to work on the "request" level, i.e. when it
# has already been turned into a basic Python datatype (usually a tuple
# consisting of the name of the function, the arguments and the keyword
# arguments) but before the request has been encoded into JSON. This
# representation is then adapted by removing fields that are not
# understood by simulators for previous versions. On the way back,
# fields that are not supplied by old simulators can be filled with
# default values that achieve the old behaviour.
#
# A disadvantage of this method is that old simulators will incur an
# extra level of indirection with each new version of the API. However,
# simulators that are up-to-date will not have any additional
# indirection.
#
# When creating a new version of the low-level API, adapt the rest of
# the mosaik code to use it. Then add a V(new)toV(old)Adapter class in
# this file which turns adapts requests conforming to the new version
# into requests conforming to the old version. Also add a version
# check to ``init_and_get_adapter`` that wraps older simulators in
# this new adapter.
from __future__ import annotations

import warnings
from typing import Any, Dict, Optional

from loguru import logger  # noqa: F401  # type: ignore
from mosaik_api_v3.types import Meta, SimId

from mosaik.exceptions import ScenarioError
from mosaik.proxies import BaseProxy, Proxy


async def init_and_get_adapter(
    base_proxy: BaseProxy,
    sid: SimId,
    sim_params: Dict[str, Any],
    explicit_version_str: Optional[str] = None,
) -> Proxy:
    """Initialize the simulator given by ``base_proxy`` (by calling its
    ``init`` function) and wrap it in a ``Proxy`` object that
    adapts it to the current version of mosaik.

    :param base_proxy: The ``BaseProxy`` for this simulator.
    :param sid: The ``SimId`` that was assigned to this simulator.
    :param kwargs: The remaining initialization arguments for the
    simulator.
    :return: The base proxy wrapped in adapters based on the API version
    returned by the simulator's ``init`` function and the meta returned
    by the init call.
    :raise ScenarioError: if there is a problem during initialization.
    """
    if explicit_version_str is not None:
        explicit_version = list(map(int, explicit_version_str.split(".")))
    else:
        explicit_version = None

    try:
        version = await base_proxy.init(sid, **sim_params)
    except ScenarioError as e:
        raise ScenarioError(
            f"There was an error during the initialization of {sid}: ", e
        )

    # version > 3.0
    if version >= [4]:
        raise ScenarioError(
            f"There was an error during the initialization of {sid}: "
            f"The API version ({'.'.join(map(str, version))}) is too new for this "
            "version of mosaik. Maybe a newer version of the mosaik package is "
            "available to be used in your scenario?"
        )
    if explicit_version and version != explicit_version:
        raise ScenarioError(
            f"The explicit version that you specified for simulator {sid} in your "
            f"SimConfig (namely {'.'.join(map(str, explicit_version))}) does not match "
            "the version that this simulator reports (namely "
            f"{'.'.join(map(str, version))})."
        )

    proxy: Proxy = base_proxy
    # Add all the adapters needed to get from the actual version of the
    # simulator to the current version.
    # Note that there are no ``elif``s here since we need to add all
    # required adapters, not just the first matching one.
    # IMPORTANT: Add new adapters at the bottom to ensure the correct
    # nesting!
    if version < [2, 2]:
        proxy = V2ToV1Adapter(proxy)
    if version < [3]:
        proxy = V3ToV2Adapter(proxy)

    # Warn the user if no explicit version was specified in the
    # SIM_CONFIG but the simulator is outdated for this version of
    # mosaik.
    if not isinstance(proxy, BaseProxy) and not explicit_version:
        warnings.warn(
            f"Simulator {sid} is using an outdated API version, namely "
            f"{'.'.join(map(str, version))}. It should still work. You "
            "can suppress this warning by adding an explicit API "
            "version for this simulator in your SimConfig."
        )

    return proxy


class Adapter(Proxy):
    _out: Proxy

    def __init__(self, out: Proxy):
        self._out = out

    @property
    def meta(self) -> Meta:
        return self._out.meta

    async def send(self, request: Any):
        return await self._out.send(request)

    async def stop(self):
        return await self._out.stop()


class V3ToV2Adapter(Adapter):
    """API changes:
    - ``init`` is now supplied with the ``time_resolution`` (as a kwarg)
      (but this is handled in BaseConnection)
    - ``step`` is now supplied with ``max_advance`` (as an arg)
    """

    async def send(self, request: Any):
        try:
            func_name, args, kwargs = request
            if func_name == "step":
                request = ("step", args[0:2], kwargs)
        except ValueError:
            pass
        return await self._out.send(request)

    @property
    def meta(self) -> Meta:
        self._out.meta.setdefault("type", "time-based")
        return self._out.meta


class V2ToV1Adapter(Adapter):
    """API changes:
    - ``setup_done`` function was added to simulators
    """

    async def send(self, request: Any):
        try:
            func_name, _args, _kwargs = request
            if func_name == "setup_done":
                return None
        except ValueError:
            pass

        return await self._out.send(request)



"""
This module provides mosaik specific exception types.
"""

from typing import Any, List, Tuple

from mosaik_api_v3 import SimId


class ScenarioError(Exception):
    """
    This exception is raised if something fails during the creation of
    a scenario.
    """


class SimulationError(Exception):
    """
    This exception is raised if a simulator cannot be started or if
    a problem arises during the execution of a simulation.
    """

    def __init__(self, msg: str, exc: Any = None):
        arg = ""
        if exc:
            orig = str(exc)
            if orig.endswith("."):
                orig = orig[:-1]
            arg += "%s: " % orig
        arg += msg
        super().__init__(arg)


class NonSerializableOutputsError(SimulationError):
    dest: SimId
    errors: List[Tuple[str, str, str, TypeError]]

    def __init__(self, dest: SimId):
        self.dest = dest
        self.errors = []

    def add_error(self, dest_eid: str, dest_attr: str, src_id: str, error: TypeError):
        self.errors.append((dest_eid, dest_attr, src_id, error))

    def __bool__(self):
        return bool(self.errors)

    def __str__(self) -> str:
        return (
            f"Errors while trying to JSON-serialize inputs for {self.dest}:\n"
            + "\n".join(
                f"- serializing output from {src} for {dest_eid}.{dest_attr}: {str(error)}"
                for dest_eid, dest_attr, src, error in self.errors
            )
            + "\nThis is likely a problem in the source simulator(s)."
        )


class SimulatorError(Exception):
    """This exception is raised if a simulator does not behave
    correctly."""

    simulator: str

    def __init__(self, simulator: str, *args: Any) -> None:
        self.simulator = simulator
        super().__init__(*args)


class DuplicateEntityIdError(SimulatorError):
    """This exception is raised if a simulator returns multiple entities
    with the same entity ID."""

    entity_id: str

    def __init__(self, simulator: str, entity_id: str, *args: Any) -> None:
        self.entity_id = entity_id
        super().__init__(simulator, *args)

    def __str__(self) -> str:
        return (
            f"Simulator {self.simulator} returned multiple entities "
            f"with entity ID '{self.entity_id}'."
        )



from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import Any, Optional, Set, Union


def tuple_add(xs: tuple[int, ...], ys: tuple[int, ...]) -> tuple[int, ...]:
    return tuple(x + y for x, y in zip(xs, ys))


@dataclass(frozen=True)
class TieredDuration:
    pre_length: int
    cutoff: int
    tiers: tuple[int, ...]

    def __init__(
        self, *tiers: int, cutoff: int | None = None, pre_length: int | None = None
    ):
        if cutoff is None:
            cutoff = len(tiers)
        if pre_length is None:
            pre_length = cutoff
        assert cutoff >= 1
        assert cutoff <= pre_length
        assert cutoff <= len(tiers)
        object.__setattr__(self, "pre_length", pre_length)
        object.__setattr__(self, "cutoff", cutoff)
        object.__setattr__(self, "tiers", tiers)

    def __len__(self) -> int:
        return len(self.tiers)

    @property
    def add(self) -> tuple[int, ...]:
        return self.tiers[0 : self.cutoff]

    @property
    def ext(self) -> tuple[int, ...]:
        return self.tiers[self.cutoff :]

    def __add__(self, other: TieredDuration) -> TieredDuration:
        assert len(self) == other.pre_length
        add = tuple_add(self.add, other.add)
        if self.cutoff >= other.cutoff:
            ext = other.ext
        else:  # self.cutoff is shorter
            ext = (
                tuple_add(self.ext + ((0,) * len(other.add)), other.add[self.cutoff :])
                + other.ext
            )
        tiers = add + ext
        cutoff = min(self.cutoff, other.cutoff)
        assert len(tiers) == len(other)
        return TieredDuration(*tiers, pre_length=self.pre_length, cutoff=cutoff)

    def __le__(self, other: TieredDuration) -> bool:
        assert len(self) == len(other)
        assert self.pre_length == other.pre_length
        joint_cutoff = min(self.cutoff, other.cutoff)
        if self.tiers[0:joint_cutoff] < other.tiers[0:joint_cutoff]:
            return True
        return self.tiers <= other.tiers and self.cutoff <= other.cutoff

    def __lt__(self, other: TieredDuration) -> bool:
        return self <= other and not self == other

    def __ge__(self, other: TieredDuration) -> bool:
        return other <= self

    def __gt__(self, other: TieredDuration) -> bool:
        return other < self

    def __repr__(self):
        return (
            f"{':'.join(map(str, self.add))}|{':'.join(map(str, self.ext))}"
            f"({self.pre_length})"
        )


@functools.total_ordering
@dataclass(frozen=True)
class TieredTime:
    tiers: tuple[int, ...]

    def __init__(self, *tiers: int):
        object.__setattr__(self, "tiers", tiers)

    def __add__(self, interval: TieredDuration) -> TieredTime:
        assert len(self.tiers) == interval.pre_length
        return TieredTime(*(tuple_add(self.tiers, interval.add) + interval.ext))

    def __lt__(self, other: TieredTime) -> bool:
        assert len(self) == len(other)
        return self.tiers < other.tiers

    def __len__(self) -> int:
        return len(self.tiers)

    @property
    def time(self) -> int:
        return self.tiers[0]

    def __repr__(self):
        return f"{':'.join(map(str, self.tiers))}"


class MinimalDurations:
    """A set of minimal TieredDurations.

    Because tiered durations are not always comparable, a set of them
    does not always have a unique minimum. (However, for a given length
    and pre-length, the number of minimal elements is at most
    min(length, pre-length).)

    This class represents a set of minimal elements. In other words, the
    elements of `durations` are pairwise incomparable. When a new
    duration is `insert`ed, `durations` will be updated to contain the
    minimal elements among all elements seen so far.
    """

    durations: Set[TieredDuration]
    """All the minimal durations. Invariant: No two durations in this
    set are comparable.
    """

    def __init__(self, duration: Optional[TieredDuration] = None):
        self.durations = set()
        if duration is not None:
            self.durations.add(duration)

    def insert(self, duration: TieredDuration) -> bool:
        """
        Insert a `TieredDuration` into the minimal duration set. If the
        new duration is bigger than any existing one, nothing happens.
        Otherwise, all exisiting durations that are bigger than the new
        one are removed and the new one is added to the set.
        (Incomparable existing duration are not touched.)

        Returns whether the set of durations was changed.
        """
        displaced_durations: Set[TieredDuration] = set()
        for existing_duration in self.durations:
            if duration < existing_duration:
                displaced_durations.add(existing_duration)
            if existing_duration <= duration:
                # Because of the invariant of `self.durations`, we don't
                # insert `duration`, and no other `existing_duration`
                # can be displaced by `duration`.
                return False
        self.durations.difference_update(displaced_durations)
        self.durations.add(duration)
        return True

    def insert_all(self, other: MinimalDurations) -> bool:
        updated = False
        for duration in other.durations:
            updated = self.insert(duration) or updated
        return updated

    def __add__(
        self, other: Union[MinimalDurations, TieredDuration]
    ) -> MinimalDurations:
        result = MinimalDurations()
        if isinstance(other, TieredDuration):
            other = MinimalDurations(other)
        for u in self.durations:
            for v in other.durations:
                result.insert(u + v)
        return result

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MinimalDurations):
            raise TypeError("cannot compare MinimalDurations to values of other types")
        return self.durations == other.durations

    def contains_zero(self) -> bool:
        return any(all(t == 0 for t in duration.tiers) for duration in self.durations)

    def earliest_sum(self, other: TieredTime) -> TieredTime:
        """Add `other` to all durations in this set and return the
        minimal result.
        """
        return min(other + duration for duration in self.durations)

    @classmethod
    def from_duration(cls, duration: TieredDuration) -> MinimalDurations:
        min_durations = MinimalDurations()
        min_durations.insert(duration)
        return min_durations



from __future__ import annotations

from typing import (
    Any,
    FrozenSet,
    Generic,
    Iterable,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from typing_extensions import TypeAlias

E = TypeVar("E")


class OutSet(Generic[E]):
    """An OutSet[E] represents all elements of the type E except for a finite number.

    In particular, `x in OutSet(elems)` is true whenever
    `x not in elems`. Set operations between OutSets and normal FrozenSets
    work as excepted.

    Note that by their nature it is not possible to iterate over an
    OutSet.
    """

    _set: FrozenSet[E]

    def __init__(self, elems: Iterable[E] = ()):
        self._set = frozenset(elems)

    def __sub__(self, other: InOrOutSet[E]) -> InOrOutSet[E]:
        if isinstance(other, OutSet):
            return other._set - self._set
        else:
            return OutSet(self._set | other)

    def __rsub__(self, rother: FrozenSet[E]) -> FrozenSet[E]:
        return rother & self._set

    def __and__(self, other: InOrOutSet[E]) -> InOrOutSet[E]:
        if isinstance(other, OutSet):
            return OutSet(self._set | other._set)
        else:
            return other - self._set

    def __rand__(self, rother: FrozenSet[E]) -> FrozenSet[E]:
        return rother - self._set

    def __or__(self, other: InOrOutSet[E]) -> OutSet[E]:
        if isinstance(other, OutSet):
            return OutSet(self._set & other._set)
        else:
            return OutSet(self._set - other)

    def __ror__(self, rother: FrozenSet[E]) -> OutSet[E]:
        return OutSet(self._set - rother)

    def __contains__(self, item: E) -> bool:
        return item not in self._set

    def __eq__(self, other: Any):
        if not isinstance(other, OutSet):
            return False
        return self._set == other._set  # type: ignore  (Pyright does not know E here)

    def __str__(self):
        return f"OutSet({{{', '.join(map(str, self._set))}}})"


InOrOutSet: TypeAlias = Union[FrozenSet[E], OutSet[E]]
"""A InOrOutSet is either a FrozenSet or an OutSet. This means
that it can represent either
- a finite number of elements of the type E or
- all but a finite number of element of the type E.

Standard set-theoretic operations (union, intersection, etc.) can still
be computed for InOrOutSets and will result in a InOrOutSet again.
"""


def parse_set_triple(
    union: Optional[InOrOutSet[E]],
    part_a: Optional[InOrOutSet[E]],
    part_b: Optional[InOrOutSet[E]],
    union_name: str = "union",
    part_a_name: str = "part_a",
    part_b_name: str = "part_b",
) -> Tuple[InOrOutSet[E], InOrOutSet[E]]:
    """Take three sets and make sure that the first is the disjoint
    union of the other two. If one of the sets is None, find the value
    for it that ensures this, if possible.
    Return the two parts.
    """
    missing_value_error = ValueError(
        f"at least two of {union_name}, {part_a_name} and {part_b_name} must be given"
    )

    if union is None:
        if part_a is not None and part_b is not None:
            union = part_a | part_b
        else:
            raise missing_value_error

    if part_a is None:
        if part_b is not None:
            part_a = union - part_b
        else:
            raise missing_value_error

    if part_b is None:
        part_b = union - part_a

    if not part_a & part_b == frozenset():
        raise ValueError(
            f"{part_a_name} ({part_a}) and {part_b_name} ({part_b}) are not disjoint"
        )

    if not union == (part_a | part_b):
        raise ValueError(
            f"{part_a_name} ({part_a}) and {part_b_name} ({part_b}) must be subsets "
            f"of {union_name} ({union}), and they must have {union_name} as their "
            f"union if both given"
        )

    return part_a, part_b


def wrap_set(set: Union[Iterable[E], OutSet[E], None]) -> Optional[InOrOutSet[E]]:
    """Wrap an iterable or OutSet, resulting in an InOrOutSet. Pass
    through None unchanged.
    """
    if set is None or isinstance(set, OutSet):
        return set
    return frozenset(set)



"""
This module provides the async interface for users to create simulation
scenarios for mosaik. For a version that handles the asynchronicity
itself (but cannot be used within an existing event loop), see
scenario.py.

The :class:`AsyncWorld` holds all necessary data for the simulation and
allows the user to start simulators. It provides a
:class:`AsyncModelFactory` (and a :class:`AsyncModelMock`) via which the
user can instantiate model instances (*entities*). The method
:meth:`AsyncWorld.run` finally starts the simulation.
"""

from __future__ import annotations

import asyncio
import contextlib
import itertools
import warnings
from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import networkx
from loguru import logger
from mosaik_api_v3.connection import RemoteException
from mosaik_api_v3.types import (
    Attr,
    CreateResult,
    EntityId,
    FullId,
    ModelDescription,
    ModelName,
    OutputRequest,
    SimId,
)
from networkx import DiGraph
from tqdm import tqdm
from typing_extensions import Literal, Self, TypeAlias, TypedDict

from mosaik import scheduler, simmanager
from mosaik.exceptions import DuplicateEntityIdError, ScenarioError, SimulationError
from mosaik.greetings_util import print_greetings
from mosaik.in_or_out_set import InOrOutSet, OutSet, parse_set_triple, wrap_set
from mosaik.internal_util import doc_link
from mosaik.proxies import Proxy
from mosaik.simmanager import MosaikConfigTotal, SimRunner
from mosaik.tiered_time import MinimalDurations, TieredDuration, TieredTime

if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison


class MosaikConfig(TypedDict, total=False):
    addr: Tuple[str, int | None]
    start_timeout: float
    stop_timeout: float


class _MosaikConfigTotal(TypedDict):
    """A total version for :cls:`MosaikConfig` for internal use."""

    addr: Tuple[str, int | None]
    start_timeout: float
    stop_timeout: float


base_config: _MosaikConfigTotal = {
    "addr": ("127.0.0.1", None),
    "start_timeout": 10,  # seconds
    "stop_timeout": 10,  # seconds
}


SENTINEL = object()
"""Sentinel for initial data call (we can't  use None as the user might
want to supply that value.)
"""


class ModelOptionals(TypedDict, total=False):
    env: Dict[str, str]
    """The environment variables to set for this simulator."""
    cwd: str
    """The current working directory for this simulator."""
    api_version: str
    """The API version of the connected simulator. Set this to suppress
    warnings about this simulator being outdated.
    """
    posix: bool
    """Whether to split the given shell command using POSIX rules.
    (Default: ``False`` on Windows, ``True`` otherwise.)
    """
    auto_terminate: bool
    """Whether to terminate this process automatically at the end of the
    simulation. (Default ``true``.)
    """


class PythonModel(ModelOptionals):
    python: str
    """The Simulator subclass for this simulator, encoded as a string
    `module_name:ClassName`."""


class ConnectModel(ModelOptionals):
    connect: str
    """The `host:port` address for this simulator."""


class CmdModel(ModelOptionals):
    cmd: str
    """The command to start this simulator. String %(python)s will be replaced
    by the python command used to start this scenario, %(addr)s will be replaced
    by the `host:port` combination to which the simulator should connect."""


SimConfig: TypeAlias = Dict[str, Union[PythonModel, ConnectModel, CmdModel]]
"""Description of all the simulators you intend to use in your
simulation.
"""


@dataclass
class SimGroup:
    parent: SimGroup | None

    @property
    def depth(self) -> int:
        if not self.parent:
            return 1
        return self.parent.depth + 1


def group_path(src: SimGroup, dest: SimGroup) -> Tuple[int, int, SimGroup]:
    src_groups = [src]
    while src.parent:
        src = src.parent
        src_groups.append(src)

    descent = 0
    while True:
        try:
            ascent = src_groups.index(dest)
            return (ascent, descent, dest)
        except ValueError:
            if not dest.parent:
                raise ValueError("no joint parent")
            dest = dest.parent
            descent += 1


def connect_interval(
    src_group: SimGroup, dest_group: SimGroup, time_shifted: int = 0, weak: int = 0
):
    """Given two `SimGroup`s, calculate a TieredInterval connecting
    simulators in these groups. The tiers will be 0, unless
    `time_shifted` or `weak` are specified, in which case the given
    values are placed in the highest and lowest tier, respectively.
    """
    ascent, _, common_group = group_path(src_group, dest_group)

    pre_length = src_group.depth
    cutoff = pre_length - ascent
    list_tiers = [0] * dest_group.depth
    if weak and not common_group.parent:
        raise ScenarioError(
            "Weak connections may only be used in groups. This is new in mosaik 3.3. "
            "For more information, see "
            f"{doc_link('scenario-definition', 'weak-connections')}."
        )
    if time_shifted:
        list_tiers[0] = time_shifted
    if weak:
        assert cutoff >= 2
        list_tiers[cutoff - 1] = weak
    return TieredDuration(*list_tiers, cutoff=cutoff, pre_length=pre_length)


class AsyncWorld:
    """
    The world holds all data required to specify and run the scenario.

    It provides a method to start a simulator process (:meth:`start`) and
    manages the simulator instances.

    You have to provide a *sim_config* which tells the world which simulators
    are available and how to start them. See :func:`mosaik.simmanager.start`
    for more details.

    *mosaik_config* can be a dict or list of key-value pairs to set addional
    parameters overriding the defaults::

        {
            'addr': ('127.0.0.1', 5555),
            'start_timeout': 2,  # seconds
            'stop_timeout': 2,   # seconds
        }

    Here, *addr* is the network address that mosaik will bind its socket to.
    *start_timeout* and *stop_timeout* specifiy a timeout (in seconds) for
    starting/stopping external simulator processes.

    If *execution_graph* is set to ``True``, an execution graph will be created
    during the simulation. This may be useful for debugging and testing. Note,
    that this increases the memory consumption and simulation time.

    We recommend that you use ``AsyncWorld`` in an ``async with`` block
    like so:

        async with AsyncWorld(SIM_CONFIG) as world:
            # call setup methods on world
            ...
            await world.run(UNTIL)

    This will ensure that the connections to all remote simulators are
    properly closed at the end of the simulation. Alternatively, you can
    use the :meth:`shutdown` method manually.
    """

    sim_config: SimConfig
    """The config dictionary that tells mosaik how to start a simulator.
    """
    config: MosaikConfigTotal
    """The config dictionary for general mosaik settings."""
    until: int  # type: ignore  # set in run
    """The time until which this simulation will run."""
    rt_factor: Optional[float]  # type: ignore  # set in run
    """The number of real-time seconds corresponding to one mosaik step."""

    time_resolution: float
    """The number of seconds that correspond to one mosaik time step in
    this situation. The default value is 1.0, meaning that one integer
    step corresponds to one second simulated time.
    """
    max_loop_iterations: int
    """The number of iterations allowed for same-time loops within one
    time step. This is checked to prevent accidental infinite loops.
    Increase this value if your same-time loops require many iterations
    to converge.
    """

    sim_progress: float
    """The progress of the entire simulation (in percent)."""
    use_cache: bool
    sims: Dict[SimId, simmanager.SimRunner]
    """A dictionary of already started simulators instances."""
    _sim_ids: Dict[ModelName, Iterator[int]]

    main_group: SimGroup
    current_group: SimGroup

    entity_graph: networkx.Graph[FullId]
    """The :class:`graph <networkx.Graph>` of related entities.
    Nodes are ``(sid, eid)`` tuples. Each note has an attribute
    *entity* with an :class:`Entity`.
    """

    tqdm: tqdm[NoReturn]  # type: ignore  # set in run
    """The tqdm progress bar for the total progress."""

    def __init__(
        self,
        sim_config: SimConfig,
        mosaik_config: Optional[MosaikConfig] = None,
        time_resolution: float = 1.0,
        debug: bool = False,
        cache: bool = True,
        max_loop_iterations: int = 100,
        skip_greetings: bool = True,
    ):
        if not skip_greetings:
            print_greetings()
        self.sim_config = sim_config

        self.config = copy(base_config)
        if mosaik_config:
            self.config.update(mosaik_config)

        self.sims = {}
        self.main_group = SimGroup(parent=None)
        self.current_group = self.main_group

        self.time_resolution = time_resolution
        self.max_loop_iterations = max_loop_iterations

        self.entity_graph = networkx.Graph()

        self.sim_progress = 0.0

        self._debug = False
        if debug:
            logger.warning(
                "You are running your simulation in debug mode. This can lead to "
                "significant slow-downs, as it will create a graph of the entire "
                "execution. Only use this mode if you intend to analyze the execution "
                "graph afterwards."
            )
            self._debug = True
            self.execution_graph: DiGraph[Tuple[SimId, TieredTime]] = DiGraph()

        # Contains ID counters for each simulator type.
        self._sim_ids = defaultdict(itertools.count)
        self.use_cache = cache

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self, exc_type: Type[Exception], exc: Exception, tb: TracebackType
    ):
        await self.shutdown()
        # Don't suppress exceptions. Later on, we might want to add
        # handling of mosaik exceptions here.
        return False

    @contextlib.contextmanager
    def group(self):
        parent_group = self.current_group
        new_group = SimGroup(parent=parent_group)
        self.current_group = new_group
        yield
        self.current_group = parent_group

    async def start(
        self,
        sim_name: str,
        sim_id: Optional[SimId] = None,
        **sim_params: Any,
    ) -> AsyncModelFactory:
        """
        Start the simulator named *sim_name* and return a
        :class:`ModelFactory` for it.
        """
        if not sim_id:
            counter = self._sim_ids[sim_name]
            sim_id = "%s-%s" % (sim_name, next(counter))
        if sim_id in self.sims:
            raise ScenarioError(
                f"A simulator with sim_id '{sim_id}' has already been started. "
                "Choose a different sim_id."
            )
        logger.info(
            'Starting "{sim_name}" as "{sim_id}" ...',
            sim_name=sim_name,
            sim_id=sim_id,
        )
        proxy = await simmanager.start(
            self, sim_name, sim_id, self.time_resolution, sim_params
        )
        # Create the ModelFactory before the SimRunner as it performs
        # some checks on the simulator's meta.
        model_factory = AsyncModelFactory(self, self.current_group, sim_id, proxy)
        self.sims[sim_id] = SimRunner(sim_id, proxy, depth=self.current_group.depth)
        if self.use_cache:
            self.sims[sim_id].outputs = {}
        return model_factory

    def connect_one(  # noqa: C901
        self,
        src: Entity,
        dest: Entity,
        src_attr: Attr,
        dest_attr: Optional[Attr] = None,
        time_shifted: Union[bool, int] = False,
        weak: bool = False,
        initial_data: Any = SENTINEL,
    ):
        if not dest_attr:
            dest_attr = src_attr

        src_sim = self.sims[src.sid]
        dest_sim = self.sims[dest.sid]

        src_port = (src.eid, src_attr)
        dest_port = (dest.eid, dest_attr)

        problems: List[str] = []

        if src_attr not in src.model_mock.output_attrs:
            problems.append("the source attribute does not exist")
        if dest_attr not in dest.model_mock.input_attrs:
            problems.append("the destination attribute does not exist")

        if (time_shifted or weak) and dest_attr in dest.model_mock.measurement_inputs:
            if initial_data is SENTINEL:
                problems.append(
                    "weak or time-shifted connection into non-trigger attribute "
                    "requires initial data"
                )
        elif initial_data is not SENTINEL:
            logger.warning(
                f"Gave initial data for connection from {src.full_id}.{src_attr} to "
                f"{dest.full_id}.{dest_attr} where it is not needed"
            )

        if problems:
            raise ScenarioError(
                f"The are problems connecting {src.full_id}.{src_attr} to "
                f"{dest.full_id}.{dest_attr}:\n- " + "\n- ".join(problems)
            )

        if (
            dest_attr in dest.model_mock.measurement_inputs
            and src_attr in src.model_mock.event_outputs
        ):
            logger.warning(
                f"A connection between the non-persistent attribute {src_attr} of "
                f"{src.sid} and the non-trigger attribute {dest_attr} of "
                f"{dest.sid} is not recommended. This might cause problems in the "
                "simulation! See also: https://mosaik.readthedocs.io/en/latest/"
                "scenario-definition.html#connecting-entities"
            )

        src_group = src.model_mock._factory._group
        dest_group = dest.model_mock._factory._group
        delay = connect_interval(src_group, dest_group, int(time_shifted), int(weak))

        dest_sim.input_delays.setdefault(src_sim, MinimalDurations()).insert(delay)

        is_pulled = src_sim.outputs is not None and src.is_persistent(src_attr)

        if src.is_persistent(src_attr) and not self.use_cache:
            dest_sim.persistent_inputs.setdefault(dest.eid, {}).setdefault(
                dest_attr, {}
            ).setdefault(src.full_id, None)

        src_sim.output_request.setdefault(src.eid, []).append(src_attr)

        if is_pulled:
            dest_sim.pulled_inputs.setdefault((src_sim, delay), set()).add(
                (src_port, dest_port)
            )
        else:
            src_sim.output_to_push.setdefault(src_port, []).append(
                (dest_sim, delay, dest_port)
            )

        src_sim.successors[dest_sim] = connect_interval(src_group, dest_group)

        if dest.triggered_by(dest_attr):
            src_sim.triggers.setdefault(src_port, []).append((dest_sim, delay))

        if initial_data is not SENTINEL:
            if is_pulled:
                assert src_sim.outputs is not None
                src_sim.outputs.setdefault(-int(time_shifted), {}).setdefault(
                    src.eid, {}
                )[src_attr] = initial_data
            else:
                dest_sim.persistent_inputs.setdefault(dest.eid, {}).setdefault(
                    dest_attr, {}
                )[src.full_id] = initial_data

        self.entity_graph.add_edge(src.full_id, dest.full_id)

    def connect_async_requests(self, src: AsyncModelFactory, dest: AsyncModelFactory):
        warnings.warn(
            "Connections using async_requests and the "
            "set_data function are deprecated and will be "
            "removed in a future release. Please use "
            "time_shifted and weak connections instead.",
            category=DeprecationWarning,
        )
        src_sim = self.sims[src._sid]
        dest_sim = self.sims[dest._sid]
        delay = connect_interval(src._group, dest._group)
        src_sim.successors[dest_sim] = delay
        src_sim.successors_to_wait_for[dest_sim] = delay
        dest_sim.input_delays.setdefault(src_sim, MinimalDurations()).insert(delay)

    def connect(
        self,
        src: Entity,
        dest: Entity,
        *attr_pairs: Union[str, Tuple[str, str]],  # type: ignore
        async_requests: bool = False,
        time_shifted: Union[bool, int] = False,
        initial_data: Dict[Attr, Any] = {},
        weak: bool = False,
    ):
        """
        Connect the *src* entity to *dest* entity.

        Establish a data-flow for each ``(src_attr, dest_attr)`` tuple in
        *attr_pairs*. If *src_attr* and *dest_attr* have the same name, you
        you can optionally only pass one of them as a single string.

        Raise a :exc:`~mosaik.exceptions.ScenarioError` if both entities share
        the same simulator instance, if at least one (src. or dest.) attribute
        in *attr_pairs* does not exist, or if the connection would introduce
        a cycle in the data-flow (e.g., A → B → C → A).

        If the *dest* simulator may make asynchronous requests to mosaik to
        query data from *src* (or set data to it), *async_requests* should be
        set to ``True`` so that the *src* simulator stays in sync with *dest*.

        An alternative to asynchronous requests are time-shifted connections.
        Their data flow is always resolved after normal connections so that
        cycles in the data-flow can be realized without introducing deadlocks.
        For such a connection *time_shifted* should be set to ``True`` and
        *initial_data* should contain a dict with input data for the first
        simulation step of the receiving simulator.

        An alternative to using async_requests to realize cyclic data-flow
        is given by the time_shifted kwarg. If set to ``True`` it marks the
        connection as cycle-closing (e.g. C → A). It must always be used with
        initial_data specifying a dict with the data sent to the destination
        simulator at the first step (e.g. *{'src_attr': value}*).
        """

        # Expand single attributes "attr" to ("attr", "attr") tuples:
        attr_pairs: Set[Tuple[Attr, Attr]] = {
            (a, a) if isinstance(a, str) else a for a in attr_pairs
        }
        errors: List[ScenarioError] = []
        for src_attr, dest_attr in attr_pairs:
            try:
                self.connect_one(
                    src,
                    dest,
                    src_attr,
                    dest_attr,
                    time_shifted=time_shifted,
                    weak=weak,
                    initial_data=initial_data.get(src_attr, SENTINEL),
                )
            except ScenarioError as e:
                errors.append(e)

        if async_requests:
            self.connect_async_requests(
                src.model_mock._factory, dest.model_mock._factory
            )

        if errors:
            raise ScenarioError(
                "While connecting entities, the following errors occurred:\n - "
                + "\n - ".join(str(e) for e in errors)
            )

        trigger: Set[Tuple[EntityId, Attr]] = set()
        for src_attr, dest_attr in attr_pairs:
            if dest.triggered_by(dest_attr):
                trigger.add((src.eid, src_attr))

        # Add relation in entity_graph
        self.entity_graph.add_edge(src.full_id, dest.full_id)

    def set_initial_event(self, sid: SimId, time: int = 0):
        """
        Set an initial step for simulator *sid* at time *time* (default=0).
        """
        sim = self.sims[sid]
        sim.next_steps = [TieredTime(time) + sim.from_world_time]

    async def get_data(
        self,
        entity_set: Iterable[Entity],
        *attributes: Attr,
    ) -> Dict[Entity, Dict[Attr, Any]]:
        """
        Get and return the values of all *attributes* for each entity of an
        *entity_set*.

        The return value is a dict mapping the entities of *entity_set* to
        dicts containing the values of each attribute in *attributes*::

            {
                Entity(...): {
                    'attr_1': 'val_1',
                    'attr_2': 'val_2',
                    ...
                },
                ...
            }
        """
        outputs_by_sim: Dict[SimId, OutputRequest] = defaultdict(dict)
        for entity in entity_set:
            outputs_by_sim[entity.sid][entity.eid] = list(attributes)

        requests = {
            sid: asyncio.create_task(self.sims[sid].get_data(outputs))
            for sid, outputs in outputs_by_sim.items()
        }
        try:
            await asyncio.gather(*requests.values())
        except ConnectionError as e:
            # Try to find the simulator that closed its connection
            for sid, task in requests.items():
                if task.exception():
                    raise SimulationError(
                        f"Simulator '{sid}' closed its connection while executing "
                        "`World.get_data()`.",
                        e,
                    ) from None
            else:
                raise RuntimeError(
                    "Could not determine which simulator closed its connection."
                )

        results_by_sim = {}
        for sid, task in requests.items():
            results_by_sim[sid] = task.result()

        results: Dict[Entity, Dict[Attr, Any]] = {}
        for entity in entity_set:
            results[entity] = results_by_sim[entity.sid][entity.eid]

        return results

    async def run(
        self,
        until: int,
        rt_factor: Optional[float] = None,
        rt_strict: bool = False,
        print_progress: Union[bool, Literal["individual"]] = False,
        lazy_stepping: bool = True,
    ):
        """Start the simulation until the simulation time *until* is
        reached.

        Before this method returns, it stops all simulators and closes
        mosaik's server socket. So this method should only be called
        once.

        :param until: The end of the simulation in mosaik time steps
            (exclusive).

        :param rt_factor: The real-time factor. If set to a number > 0,
            the simulation will run in real-time mode. A real-time
            factor of 1. means that 1 second in simulated time takes
            1 second in real time. An real-time factor of 0.5 will let
            the simulation run twice as fast as real time. For correct
            behavior of the real-time factor, the time resolution of the
            scenario has to be set adequately (the default is 1 second).

        :param rt_strict: If the simulators are too slow for the
            real-time factor you chose, mosaik will only print a warning
            by default. In order to raise a :exc:`RuntimeError` instead,
            you can set *rt_strict* to ``True``.

        :param print_progress: Whether progress bars are printed while
            the simulation is running. The default is to print one bar
            representing the global progress of the simulation. You can
            also set the value to ``'individual'`` to get one bar per
            simulator in your simulation (in addition to the global
            one). A value of ``False`` turns off the progress bars
            completely.

            The progress bars use
            `tqdm <https://pypi.org/project/tqdm/>`_; see their
            documentation on how to write to the console without
            interfering with the bars.

        :param lazy_stepping: Whether to prevent simulators from running
            ahead of their successors by more than one step. If
            ``False`` a simulator always steps as long all input is
            provided. This might decrease the simulation time but
            increase the memory consumption.
        """
        if hasattr(self, "until"):
            raise RuntimeError(
                "Simulation has already been run and can only be run once for a World "
                "instance."
            )

        # Check if a simulator is not connected to anything:
        # TODO: Rebuild this test without df_graph (or maybe check for
        # connectedness instead).

        # Creating the topological ranking will ensure that there are no
        # cycles in the dataflow graph that are not resolved using
        # time-shifted or weak connections.
        self.ensure_no_dataflow_cycles()

        self.cache_triggering_ancestors()

        logger.info("Starting simulation.")
        # 11 is the length of "Total: 100%"
        max_sim_id_len = max(max(len(str(sid)) for sid in self.sims), 11)
        until_len = len(str(until))
        self.tqdm = tqdm(
            total=until,
            disable=not print_progress,
            colour="green",
            bar_format=(
                None
                if print_progress != "individual"
                else (
                    "Total:%s {percentage:3.0f}%% |{bar}| %s{elapsed}<{remaining}"
                    % (" " * (max_sim_id_len - 11), "  " * until_len)
                )
            ),
            unit="steps",
        )
        for sid, sim in self.sims.items():
            sim.tqdm = tqdm(
                total=until,
                desc=sid,
                bar_format=(
                    "{desc:>%i} |{bar}| {n_fmt:>%i}/{total_fmt}{postfix:10}"
                    % (max_sim_id_len, until_len)
                ),
                leave=False,
                disable=print_progress != "individual",
            )
        import mosaik._debug as dbg  # always import, enable when requested

        if self._debug:
            dbg.enable()
        success = False
        try:
            await scheduler.run(self, until, rt_factor, rt_strict, lazy_stepping)
            success = True
        except KeyboardInterrupt:
            logger.info("Simulation canceled. Terminating ...")
        except RemoteException as exc:
            logger.error(
                f"Simulator {exc.source} aborted the simulation with error "
                f"{exc.remote_type}({exc.remote_msg})."
                + (
                    " Further information provided:\n" + "\n".join(exc.further_args)
                    if exc.further_args
                    else ""
                )
            )
        finally:
            for sid, sim in self.sims.items():
                sim.tqdm.close()
            self.tqdm.close()
            if self._debug:
                dbg.disable()
            if success:
                logger.info("Simulation finished successfully.")

    def cache_triggering_ancestors(self):
        """Collects the ancestors of each simulator and stores them in
        the respective simulator object.
        """
        # See ``ensure_no_dataflow_cycles`` for an explanation of this
        # algorithm
        dirty: Set[SimRunner] = set()
        for sim in self.sims.values():
            for port_triggers in sim.triggers.values():
                for dest_sim, delay in port_triggers:
                    dest_sim.triggering_ancestors.setdefault(
                        sim, MinimalDurations()
                    ).insert(delay)
                    dirty.add(dest_sim)
        while dirty:
            sim = dirty.pop()
            for port_triggers in sim.triggers.values():
                for dest_sim, mid_to_dest in port_triggers:
                    for src_sim, src_to_mid in sim.triggering_ancestors.items():
                        src_to_dest = src_to_mid + mid_to_dest
                        was_updated = dest_sim.triggering_ancestors.setdefault(
                            src_sim, MinimalDurations()
                        ).insert_all(src_to_dest)
                        if was_updated:
                            dirty.add(dest_sim)
        return

    def ensure_no_dataflow_cycles(self):
        """Make sure that there is no cyclic dataflow with 0 total
        delay. Raise an exception with one such cycle otherwise.
        """

        # Short note on terminology: ancestors and descendants any
        # number of steps removed; predecessors and successors are one
        # step removed (i.e. they're *direct* ancestors/descendants)

        dirty: Set[SimRunner] = set(self.sims.values())
        """Sims that have changed descendants and thus require
        recalculation
        """
        sim_descs: Dict[SimRunner, Dict[SimRunner, MinPath]] = {
            sim: {} for sim in self.sims.values()
        }
        """For each SimRunner, all its descendants that have been found
        so far with the shortest delay to them and the path that
        exhibits this delay.
        """

        # At the start, enter each sim as a descendant of all its
        # (direct) predecessors
        for sim in self.sims.values():
            for pred, delay in sim.input_delays.items():
                min_durations = MinimalDurations()
                min_durations.insert_all(delay)
                sim_descs[pred][sim] = {
                    "delays": min_durations,
                    "path": [pred, sim],
                }
        while dirty:
            mid_sim = dirty.pop()
            # This sim has had updates to its descendants since we last
            # processed it. These changes are pushed to its predecessors
            for src_sim, src_to_mid_delays in mid_sim.input_delays.items():
                for dest_sim, mid_to_dest in sim_descs[mid_sim].items():
                    mid_to_dest_delays = mid_to_dest["delays"]
                    src_to_dest_delays = src_to_mid_delays + mid_to_dest_delays
                    was_updated = (
                        sim_descs[src_sim]
                        .setdefault(
                            dest_sim,
                            MinPath(
                                delays=MinimalDurations(),
                                path=[src_sim, dest_sim],
                            ),
                        )["delays"]
                        .insert_all(src_to_dest_delays)
                    )
                    if was_updated:
                        # In this case we need to update the
                        # predecessor's shortest path to the descendant
                        # and reprocess the predecessor
                        sim_descs[src_sim][dest_sim]["path"] = [
                            src_sim,
                            *mid_to_dest["path"],
                        ]
                        dirty.add(src_sim)

        # We need to raise an error if any sim has itself as a
        # descendant with a delay of 0
        for sim, descs in sim_descs.items():
            if sim not in descs:
                continue
            min_path = descs[sim]
            if min_path["delays"].contains_zero():
                raise ScenarioError(
                    f"Your scenario contains cycles, for example: {min_path['path']}."
                )

    async def shutdown(self):
        """
        Shut-down all simulators and close the server socket.
        """
        for sim in self.sims.values():
            await sim.stop()


class MinPath(TypedDict):
    delays: MinimalDurations
    path: List[SimRunner]


if TYPE_CHECKING:
    T = TypeVar("T", bound=SupportsRichComparison)


def update_min(a: T | None, b: T) -> T | None:
    if a is None:
        return b
    if a <= b:  # type: ignore
        return None
    return b


MOSAIK_METHODS = {
    "init",
    "create",
    "setup_done",
    "step",
    "get_data",
    "finalize",
    "stop",
}


FULL_ID = "%s.%s"


class ExtraMethodsProxy:
    _sim_id: SimId
    _methods: Set[str]

    def __init__(self, sim_id: SimId):
        self._sim_id = sim_id
        self._methods = set()

    def _add_extra_method(self, name: str, wrapper: Callable[..., Any]) -> None:
        setattr(self, name, wrapper)
        self._methods.add(name)

    def __iter__(self):
        return iter(self._methods)

    def __getattr__(self, name: str) -> Callable[..., Any]:
        raise ScenarioError(f"`{name}` is not an extra method on '{self._sim_id}'")


class AsyncModelFactory:
    """
    This is a facade for a simulator *sim* that allows the user to create
    new model instances (entities) within that simulator.

    For every model that a simulator publicly exposes, the ``ModelFactory``
    provides a :class:`ModelMock` attribute that actually creates the entities.

    If you access an attribute that is not a model or if the model is not
    marked as *public*, an :exc:`~mosaik.exceptions.ScenarioError` is raised.
    """

    type: Literal["event-based", "time-based", "hybrid"]
    models: Dict[ModelName, AsyncModelMock]

    def __init__(  # noqa: C901
        self, world: AsyncWorld, group: SimGroup, sid: SimId, proxy: Proxy
    ):
        self.meta = proxy.meta
        self._world = world
        self._group = group
        self._proxy = proxy
        self._sid = sid
        self.call = ExtraMethodsProxy(sid)

        if "type" not in proxy.meta:
            raise ScenarioError(
                'The simulator is missing a type specification ("time-based", '
                '"event-based" or "hybrid"). This is required starting from API '
                "version 3."
            )
        self.type = proxy.meta["type"]
        if self.type not in ["time-based", "event-based", "hybrid"]:
            raise ScenarioError(
                f"The type '{self.type}' is not a valid type. (It should be one of"
                "'time-based', 'event-based' and 'hybrid'.) Please check for typos "
                f"in your simulator's init function and meta."
            )

        self.models = {}
        for model, props in self.meta["models"].items():
            if model in MOSAIK_METHODS:
                raise ScenarioError(
                    f"Simulator {sid} uses an illegal model name: {model}. This name "
                    "is already the name of a mosaik API method."
                )
            self.models[model] = AsyncModelMock(self._world, self, model, self._proxy)
            # Make public models accessible
            if props.get("public", True):
                setattr(self, model, self.models[model])

        # Bind extra_methods to this instance:
        for meth_name in self.meta.get("extra_methods", []):
            if meth_name in MOSAIK_METHODS:
                raise ScenarioError(
                    f"Simulator {sid} uses an illegal name for an extra method: "
                    f'"{meth_name}". This is already the name of a mosaik API method.'
                )
            if meth_name in self.models.keys():
                raise ScenarioError(
                    f"Simulator {sid} uses an illegal name for an extra method: "
                    f'"{meth_name}". This is already the name of a model of this '
                    "simulator."
                )

            # We need get_wrapper() in order to avoid problems with
            # scoping of the name `meth_name`. Without it, `meth_name`
            # would be the same for all wrappers.
            def get_wrapper(connection: Proxy, meth_name: str) -> Callable[..., Any]:
                async def wrapper(*args: Any, **kwargs: Any):
                    return await connection.send([meth_name, args, kwargs])

                wrapper.__name__ = meth_name
                return wrapper

            self.call._add_extra_method(meth_name, get_wrapper(proxy, meth_name))

    def __getattr__(self, name: str) -> AsyncModelMock:
        # Implemented in order to improve error messages.
        models = self.meta["models"]
        if name in models:
            raise AttributeError(f'Model "{name}" is not public.')
        else:
            raise AttributeError(
                f'Model factory for "{self._sid}" has no model and no function '
                f'"{name}".'
            )


def parse_attrs(
    model_desc: ModelDescription, type: Literal["time-based", "event-based", "hybrid"]
) -> Tuple[InOrOutSet[Attr], InOrOutSet[Attr], InOrOutSet[Attr], InOrOutSet[Attr]]:
    """Parse the attrs and their trigger/persistent state.

    The guiding principle is this: The user can specify as little
    information as possible and the rest will be inferred, but
    inconsistent information will lead to an error.

    If attrs, trigger and non-trigger are all given, trigger and
    non-trigger must form a partition of attrs. If only two are given,
    the third in inferred, provided this can be done in such a way that
    trigger and non-trigger form a partition of attrs. If
    any_inputs=True, the set of all possible attrs is used instead of
    the ones specified in attrs. If only attrs is given, a default
    is chosen for the others, based on the type of the simulator.

    The same applieds to attrs, persistent and non-persistent, except
    that any_inputs is not considered (as these are outputs).

    :param model_desc: The `ModelDescription` to parse
    :param type: The simulator's type (for setting default attribute
        types).
    :return: A four-tuple of :class:`InOrOutSet`, giving the
        measurement inputs, event inputs, measurement outputs, and event
        outputs.
    :raises ValueError: if the information is insufficient or
        inconsistent
    """
    error_template = (
        "%s simulators may not specify %s attrs (use a hybrid simulator, instead, "
        "if you need both types of %s attributes), and they must list all their "
        "attrs as %s if that key is present"
    )

    if model_desc.get("any_inputs", False):
        inputs: Optional[InOrOutSet[Attr]] = OutSet()
    else:
        inputs = wrap_set(model_desc.get("attrs"))
    empty: FrozenSet[Attr] = frozenset()
    if type == "time-based":
        default_measurements = None
        default_events = empty
    elif type == "event-based":
        default_measurements = empty
        default_events = None
    elif type == "hybrid":
        default_measurements = None if "trigger" in model_desc else inputs
        default_events = None
    measurement_inputs = wrap_set(model_desc.get("non-trigger", default_measurements))
    event_inputs = wrap_set(model_desc.get("trigger", default_events))
    measurement_inputs, event_inputs = parse_set_triple(
        inputs, measurement_inputs, event_inputs, "attrs", "non-trigger", "trigger"
    )
    if type == "time-based" and event_inputs != frozenset():
        raise ValueError(
            error_template % ("time-based", "trigger", "input", "non-trigger")
        )
    if type == "event-based" and measurement_inputs != frozenset():
        raise ValueError(
            error_template % ("event-based", "non-trigger", "inpus", "trigger")
        )

    outputs = wrap_set(model_desc.get("attrs"))
    default_measurements = empty if type == "event-based" else None
    measurement_outputs = wrap_set(model_desc.get("persistent", default_measurements))
    default_events = None if type == "event-based" else empty
    event_outputs = wrap_set(model_desc.get("non-persistent", default_events))
    measurement_outputs, event_outputs = parse_set_triple(
        outputs,
        measurement_outputs,
        event_outputs,
        "attrs",
        "persistent",
        "non-persistent",
    )
    if type == "time-based" and event_outputs != frozenset():
        raise ValueError(
            error_template % ("time-based", "non-persistent", "output", "persistent")
        )
    if type == "event-based" and measurement_outputs != frozenset():
        raise ValueError(
            error_template % ("event-based", "persistent", "output", "non-persistent")
        )

    return measurement_inputs, event_inputs, measurement_outputs, event_outputs


class AsyncModelMock(object):
    """
    Instances of this class are exposed as attributes of
    :class:`ModelFactory` and allow the instantiation of simulator models.

    You can *call* an instance of this class to create exactly one entity:
    ``sim.ModelName(x=23)``. Alternatively, you can use the :meth:`create()`
    method to create multiple entities with the same set of parameters at once:
    ``sim.ModelName.create(3, x=23)``.
    """

    name: ModelName
    _world: AsyncWorld
    _factory: AsyncModelFactory
    _proxy: Proxy
    params: FrozenSet[str]
    event_inputs: InOrOutSet[Attr]
    measurement_inputs: InOrOutSet[Attr]
    event_outputs: InOrOutSet[Attr]
    measurement_outputs: InOrOutSet[Attr]

    def __init__(
        self,
        world: AsyncWorld,
        factory: AsyncModelFactory,
        model: ModelName,
        proxy: Proxy,
    ):
        self._world = world
        self._factory = factory
        self.name = model
        self._proxy = proxy
        model_desc = proxy.meta["models"][model]
        self.params = frozenset(model_desc.get("params", []))

        try:
            (
                self.measurement_inputs,
                self.event_inputs,
                self.measurement_outputs,
                self.event_outputs,
            ) = parse_attrs(model_desc, self._factory.type)
        except ValueError as e:
            raise ValueError(
                f"while parsing the model description of model {model} of the "
                f"simulator {factory._sid}: {e}"
            )

    @property
    def input_attrs(self) -> InOrOutSet[Attr]:
        return self.event_inputs | self.measurement_inputs

    @property
    def output_attrs(self) -> InOrOutSet[Attr]:
        return self.event_outputs | self.measurement_outputs

    async def __call__(self, **model_params: Any):
        """
        Call :meth:`create()` to instantiate one model.
        """
        return (await self.create(1, **model_params))[0]

    async def create(self, num: int, **model_params: Any):
        """
        Create *num* entities with the specified *model_params* and return
        a list with the entity dicts.

        The returned list of entities is the same as returned by
        :meth:`mosaik_api_v3.Simulator.create()`, but the simulator is prepended
        to every entity ID to make them globally unique.
        """
        self._check_params(**model_params)

        entities = await self._proxy.send(["create", (num, self.name), model_params])
        assert (
            len(entities) == num
        ), f"{num} entities were requested but {len(entities)} were created."

        return self._make_entities(entities, assert_type=self.name)

    def _check_params(self, **model_params: Any):
        unexpected_params = set(model_params.keys()).difference(self.params)
        if unexpected_params:
            sep = "', '"
            raise TypeError(
                "create() got unexpected keyword arguments: '"
                f"{sep.join(unexpected_params)}'"
            )

    def _make_entities(
        self, entity_dicts: List[CreateResult], assert_type: Optional[ModelName] = None
    ) -> List[Entity]:
        """
        Recursively create lists of :class:`Entity` instance from a list
        of *entity_dicts*.
        """
        sid = self._factory._sid
        entity_graph = self._world.entity_graph

        entity_set: List[Entity] = []
        for e in entity_dicts:
            self._assert_model_type(assert_type, e)

            children = e.get("children")
            if children is not None:
                children = self._make_entities(children)
            model = self._factory.models[e["type"]]
            entity = Entity(
                sid, e["eid"], self.name, model, children, e.get("extra_info")
            )

            entity_set.append(entity)
            # Test for the `created` flag. (There might be other nodes
            # in the graph that arise from edges between related
            # entities where the second entity has not yet been
            # processed.)
            if entity_graph.nodes.get(entity.full_id, {}).get("created", False):
                raise DuplicateEntityIdError(sid, entity.eid)
            entity_graph.add_node(entity.full_id, sid=sid, type=e["type"], created=True)
            for rel in e.get("rel", []):
                entity_graph.add_edge(entity.full_id, FULL_ID % (sid, rel))

        return entity_set

    def _assert_model_type(
        self, assert_type: Optional[ModelName], e: CreateResult
    ) -> None:
        """Assert that entity ``e`` has entity type ``assert_type``, or
        any valid model type of the simulator if ``assert_type`` is
        ``None``.
        """
        if assert_type is not None:
            assert e["type"] == assert_type, (
                f'Entity "{e["eid"]}" has the wrong type: "{e["type"]}"; '
                f'"{assert_type}" required.'
            )
        else:
            assert e["type"] in self._proxy.meta["models"], (
                f'Type "{e["type"]}" of entity "{e["eid"]}" not found in sim\'s meta '
                "data."
            )


class Entity(object):
    """
    An entity represents an instance of a simulation model within mosaik.
    """

    __slots__ = ["sid", "eid", "sim_name", "model_mock", "children", "extra_info"]
    sid: SimId
    """The ID of the simulator this entity belongs to."""
    eid: EntityId
    """The entity's ID."""
    sim_name: str
    """The entity's simulator name."""
    model_mock: AsyncModelMock
    """The entity's type (or class)."""
    children: List[Entity]
    """An entity set containing subordinate entities."""
    extra_info: Any

    def __init__(
        self,
        sid: SimId,
        eid: EntityId,
        sim_name: str,
        model_mock: AsyncModelMock,
        children: Optional[Iterable[Entity]],
        extra_info: Any = None,
    ):
        self.sid = sid
        self.eid = eid
        self.sim_name = sim_name
        self.model_mock = model_mock
        self.children = list(children) if children is not None else []
        self.extra_info = extra_info

    @property
    def type(self) -> ModelName:
        return self.model_mock.name

    @property
    def model(self) -> ModelName:
        return self.model_mock.name

    @property
    def full_id(self) -> FullId:
        """
        Full, globally unique entity id ``sid.eid``.
        """
        return FULL_ID % (self.sid, self.eid)

    def triggered_by(self, attr: Attr) -> bool:
        return attr in self.model_mock.event_inputs

    def is_persistent(self, attr: Attr) -> bool:
        return attr in self.model_mock.measurement_outputs

    def __str__(self):
        return (
            f"{self.__class__.__name__}({self.full_id!r}, "
            f"model={self.model_mock.name!r})"
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(full_id={self.full_id!r}, "
            f"model_mock={self.model_mock!r}, children={self.children!r})"
        )



import platform
from typing import Callable, Dict

from typing_extensions import TypeVar

from mosaik._version import version

K = TypeVar("K")
V = TypeVar("V")


def merge_existing(
    merger: Callable[[V, V], V],
    target: Dict[K, V],
    other: Dict[K, V],
) -> Dict[K, V]:
    """Merge the values from ``other`` which correspond to keys that
    already exists in ``target`` into ``target`` (and return
    ``target``). The function ``merger`` will be used to combine the
    existing value and the new value from ``other``; this will often be
    another call to :fun:`merge_existing` to extend merging to several
    dict levels. If a key only exists in ``target``, its value will be
    kept. If a key only exists in ``other`` it will not appear in the
    result.
    """
    for k, v in target.items():
        if k in other:
            target[k] = merger(v, other[k])
    return target


def merge_all(
    merger: Callable[[V, V], V],
    target: Dict[K, V],
    other: Dict[K, V],
) -> Dict[K, V]:
    """Merge dict ``other`` into ``target``, which will be modified
    (and returned). If a key exists in both ``other`` and ``target``,
    the function ``merger`` will be called on the corresponding values
    to combine them. (In case of nested dictionaries, this will often
    be another call to :fun:`merge_all`). If only one of ``target``
    or ``other`` contains a key, the corresponding value will be used
    in the result.
    """
    for k, v in other.items():
        if k in target:
            target[k] = merger(target[k], v)
        else:
            target[k] = v
    return target


def doc_link(page: str, anchor: str) -> str:
    return f"https://mosaik.readthedocs.io/en/{version}/{page}.html#{anchor}"


def get_python_version():
    return platform.python_version()


def get_os():
    return platform.platform()



"""
The simulation manager is responsible for starting simulation processes and
shutting them down. It also manages the communication between mosaik and the
processes.

It is able to start pure Python simulators in-process (by importing and
instantiating them), to start external simulation processes and to connect to
already running simulators and manage access to them.
"""

from __future__ import annotations

import asyncio
import collections
import heapq as hq
import importlib
import itertools
import os
import platform
import shlex
import subprocess
import sys
from ast import literal_eval
from json import JSONEncoder
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    NoReturn,
    Optional,
    OrderedDict,
    Set,
    Tuple,
    Union,
    cast,
)

import mosaik_api_v3
import tqdm
from loguru import logger
from mosaik_api_v3.connection import Channel
from mosaik_api_v3.types import (
    Attr,
    EntityId,
    FullId,
    InputData,
    OutputData,
    OutputRequest,
    SimId,
    Time,
)
from typing_extensions import Literal, TypeAlias, TypedDict

from mosaik.adapters import init_and_get_adapter
from mosaik.exceptions import (
    NonSerializableOutputsError,
    ScenarioError,
    SimulationError,
)
from mosaik.progress import Progress
from mosaik.proxies import BaseProxy, LocalProxy, Proxy, RemoteProxy
from mosaik.tiered_time import MinimalDurations, TieredDuration, TieredTime

if "Windows" in platform.system():
    from subprocess import CREATE_NEW_CONSOLE  # type: ignore (only Windows)

if TYPE_CHECKING:
    from mosaik.async_scenario import AsyncWorld, CmdModel, ConnectModel, PythonModel

FULL_ID_SEP = "."  # Separator for full entity IDs
FULL_ID = "%s.%s"  # Template for full entity IDs ('sid.eid')


class MosaikConfigTotal(TypedDict):
    """A total version for :cls:`MosaikConfig` for internal use."""

    addr: Tuple[str, int | None]
    start_timeout: float
    stop_timeout: float


async def start(
    world: AsyncWorld,
    sim_name: str,
    sim_id: SimId,
    time_resolution: float,
    sim_params: Dict[str, Any],
) -> Proxy:
    """
    Start the simulator *sim_name* based on the configuration im
    *world.sim_config*, give it the ID *sim_id* and pass the time_resolution
    and the parameters of the dict *sim_params* to it.

    The sim config is a dictionary with one entry for every simulator. The
    entry itself tells mosaik how to start the simulator::

        {
            'ExampleSimA': {
                'python': 'example_sim.mosaik:ExampleSim',
            },
            'ExampleSimB': {
                'cmd': 'example_sim %(addr)s',
                'cwd': '.',
                'auto_terminate': True,
            },
            'ExampleSimC': {
                'connect': 'host:port',
            },
        }

    *ExampleSimA* is a pure Python simulator. Mosaik will import the module
    ``example_sim.mosaik`` and instantiate the class ``ExampleSim`` to start
    the simulator.

    *ExampleSimB* would be started by executing the command *example_sim* and
    passing the network address of mosaik das command line argument. You can
    optionally specify a *current working directory*. It defaults to ``.``.
    You can also specify whether mosaik should attempt to terminate the
    process at the end of the simulation using the *auto_terminate*
    field. This defaults to ``True``.

    *ExampleSimC* can not be started by mosaik, so mosaik tries to connect to
    it.

    *time_resolution* (in seconds) is a global scenario parameter, which tells
    the simulators what the integer time step means in seconds. Its default
    value is 1., meaning one integer step corresponds to one second simulated
    time.

    The function returns a :class:`mosaik_api_v3.Simulator` instance.

    It raises a :exc:`~mosaik.exceptions.SimulationError` if the simulator
    could not be started.

    Return a :class:`SimProxy` instance.
    """
    try:
        sim_config = world.sim_config[sim_name]
    except KeyError:
        raise ScenarioError(
            'Simulator "%s" could not be started: Not found ' "in sim_config" % sim_name
        )

    # Try available starters in that order and raise an error if none of them
    # matches. Default starters are:
    # - python: start_inproc
    # - cmd: start_proc
    # - connect: start_connect
    starters = StarterCollection()

    for sim_type, starter in starters.items():
        if sim_type in sim_config:
            proxy = await starter(
                world.config, sim_name, sim_config, MosaikRemote(world, sim_id)
            )
            try:
                proxy = await asyncio.wait_for(
                    init_and_get_adapter(
                        proxy,
                        sim_id,
                        {"time_resolution": time_resolution, **sim_params},
                        explicit_version_str=sim_config.get("api_version"),
                    ),
                    world.config["start_timeout"],
                )
                return proxy
            except asyncio.IncompleteReadError:
                await proxy.stop()
                raise SystemExit(
                    f'Simulator "{sim_name}" closed its connection during the init() '
                    "call."
                )
            except asyncio.TimeoutError:
                await proxy.stop()
                raise SystemExit(
                    f'Simulator "{sim_name}" did not reply to the init() call in time.'
                )
    else:
        raise ScenarioError(
            f'Simulator "{sim_name}" could not be started: Invalid configuration'
        )


async def start_inproc(
    mosaik_config: MosaikConfigTotal,
    sim_name: str,
    sim_config: PythonModel,
    mosaik_remote: MosaikRemote,
) -> BaseProxy:
    """
    Import and instantiate the Python simulator *sim_name* based on its
    config entry *sim_config*.

    Return a :class:`LocalProcess` instance.

    Raise a :exc:`~mosaik.exceptions.ScenarioError` if the simulator cannot be
    instantiated.
    """
    try:
        mod_name, cls_name = sim_config["python"].split(":")
        mod = importlib.import_module(mod_name)
        cls = getattr(mod, cls_name)
    except (AttributeError, ImportError, KeyError, ValueError) as err:
        detail_msgs = {
            ValueError: 'Malformed Python class name: Expected "module:Class"',
            ModuleNotFoundError: "Could not import module: %s" % err.args[0],
            AttributeError: "Class not found in module",
            ImportError: f"Error importing the requested class: {err.args[0]}",
            KeyError: "'python' key not found in sim_config. "
            "(This is an error in mosaik, please report it.)",
        }
        details = detail_msgs[type(err)]
        origerr = err.args[0]
        raise ScenarioError(
            'Simulator "%s" could not be started: %s --> %s'
            % (sim_name, details, origerr)
        ) from None
    sim = cls()

    if int(mosaik_api_v3.__version__.split(".")[0]) < 3:
        raise ScenarioError("Mosaik 3 requires mosaik_api_v3 or newer.")

    return LocalProxy(sim, mosaik_remote)


async def start_proc(
    mosaik_config: MosaikConfigTotal,
    sim_name: str,
    sim_config: CmdModel,
    mosaik_remote: MosaikRemote,
) -> BaseProxy:
    """
    Start a new process for simulator *sim_name* based on its config entry
    *sim_config*.

    Return a :class:`RemoteProcess` instance.

    Raise a :exc:`~mosaik.exceptions.ScenarioError` if the simulator cannot be
    instantiated.
    """
    channel_future: asyncio.Future[Channel] = asyncio.Future()

    async def on_connect(r: asyncio.StreamReader, w: asyncio.StreamWriter):
        channel_future.set_result(Channel(r, w, name=sim_name))

    server = await asyncio.start_server(on_connect, *mosaik_config["addr"])
    try:
        actual_addr = server.sockets[0].getsockname()

        replacements = {
            "addr": "%s:%s" % actual_addr,
            "python": sys.executable,
        }
        cmd = sim_config["cmd"] % replacements
        posix = sim_config.pop("posix", os.name != "nt")
        cmd = shlex.split(cmd, posix=bool(posix))
        cwd = sim_config.get("cwd", ".")

        # Make a copy of the current env vars dictionary and update it with the
        # user provided values (or an empty dict as a default):
        env = dict(os.environ)
        env.update(sim_config.get("env", {}))

        # CREATE_NEW_CONSOLE constant for subprocess is only available on Windows
        creationflags: int = 0
        new_console = sim_config.get("new_console", False)
        if new_console:
            if "Windows" in platform.system():
                creationflags = cast(int, CREATE_NEW_CONSOLE)  # type: ignore
            else:
                logger.warning(
                    f'Simulator "{sim_name}" could not be started in a new console: '
                    "Only available on Windows"
                )

        try:
            proc = subprocess.Popen(
                cmd,
                bufsize=1,
                cwd=cwd,
                universal_newlines=True,
                env=env,  # pass the new env dict to the sub process
                creationflags=creationflags,
            )
        except (FileNotFoundError, NotADirectoryError) as e:
            # This distinction has to be made due to a change in python 3.8.0.
            # It might become unecessary for future releases supporting
            # python >= 3.8 only.
            if str(e).count(":") == 2:
                eout = e.args[1]
            else:
                eout = str(e).split("] ")[1]
            raise ScenarioError(
                f'Simulator "{sim_name}" could not be started: {eout}'
            ) from None

        try:
            channel = await asyncio.wait_for(
                channel_future, timeout=mosaik_config["start_timeout"]
            )
            return RemoteProxy(
                channel,
                mosaik_remote,
                process=(proc, sim_config.get("auto_terminate", True)),
            )
        except asyncio.TimeoutError:
            if sim_config.get("auto_terminate", True):
                proc.terminate()
            raise SimulationError(
                f'Simulator "{sim_name}" did not connect to mosaik in time.'
            )
    finally:
        server.close()


async def start_connect(
    mosaik_config: MosaikConfigTotal,
    sim_name: str,
    sim_config: ConnectModel,
    mosaik_remote: MosaikRemote,
) -> BaseProxy:
    """
    Connect to the already running simulator *sim_name* based on its config
    entry *sim_config*.

    Return a :class:`RemoteProcess` instance.

    Raise a :exc:`~mosaik.exceptions.ScenarioError` if the simulator cannot be
    instantiated.
    """
    addr = sim_config["connect"]
    try:
        host, port = addr.strip().split(":")
        addr = (host, int(port))
    except ValueError:
        raise ScenarioError(
            f'Simulator "{sim_name}" could not be started: Could not parse address '
            f'"{sim_config["connect"]}"'
        ) from None

    try:
        reader, writer = await asyncio.open_connection(host, port)
    except (ConnectionError, OSError):
        raise SimulationError(
            f'Simulator "{sim_name}" could not be started: Could not connect to '
            f'"{sim_config["connect"]}"'
        )
    return RemoteProxy(Channel(reader, writer, name=sim_name), mosaik_remote)


Port: TypeAlias = Tuple[EntityId, Attr]
"""Pair of an entity ID and an attribute of that entity"""


class SimRunner:
    """
    Handler for an external simulator.

    It stores its simulation state and own the proxy object to the external
    simulator.
    """

    sid: SimId
    """This simulator's ID."""
    type: Literal["time-based", "event-based", "hybrid"]
    supports_set_events: bool

    _proxy: Proxy
    """The actual proxy for this simulator."""

    # Connection setup
    input_delays: Dict[SimRunner, MinimalDurations]
    """For each simulator that provides data to this simulator, the
    minimum over all input delays. This is used while waiting for
    dependencies.
    """
    # TODO: Saving the minimal durations here might actually be wrong.
    # We probably want to save *all* triggering durations.
    triggers: Dict[Port, List[Tuple[SimRunner, TieredDuration]]]
    """For each port of this simulator, the simulators that are
    triggered by output on that port and the delay accrued along that
    edge.
    """
    successors: Dict[SimRunner, TieredDuration]
    """The immediate successors of this simulator. This is used when
    lazy stepping to ensure that we don't step ahead too far. Therefore,
    the duration is only used as an adapter, and will always have all
    tiers 0. (Thus, we don't need `MinimalDurations` here.)
    """
    successors_to_wait_for: Dict[SimRunner, TieredDuration]
    """The immediate successors that we always need to wait for (due
    to async requests.) The duration only serves as an adapter (so we
    don't need `MinimalDurations` here.)
    """
    triggering_ancestors: Dict[SimRunner, MinimalDurations]
    """An iterable of this sim's ancestors that can trigger a step of
    this simulator. The second component specifies the least amount of
    time that output from the ancestor needs to reach us.
    """
    pulled_inputs: Dict[Tuple[SimRunner, TieredDuration], Set[Tuple[Port, Port]]]
    """Output to pull in whenever this simulator performs a step.
    The keys are the source SimRunner and the time shift, the values
    are the source and destination entity-attribute pairs.
    """
    output_to_push: Dict[Port, List[Tuple[SimRunner, TieredDuration, Port]]]
    """This lists those connections that use the timed_input_buffer.
    The keys are the entity-attribute pairs of this simulator with
    the corresponding list of simulator-time-entity-attribute triples
    describing the destinations for that data and the time-shift
    occuring along the connection.
    """

    to_world_time: TieredDuration
    from_world_time: TieredDuration

    output_request: OutputRequest

    inputs_from_set_data: InputData
    """Inputs received via `set_data`."""
    persistent_inputs: InputData
    """Memory of previous inputs for persistent attributes."""
    timed_input_buffer: TimedInputBuffer
    """Inputs for this simulator."""

    rt_start: float  # type: ignore  # set at start of sim_process
    """The real time when this simulator started (as returned by
    `perf_counter()`."""
    started: bool

    next_steps: List[TieredTime]
    """The scheduled next steps this simulator will take, organized as a heap.
    Once the immediate next step has been chosen (and the `has_next_step` event
    has been triggered), the step is moved to `next_step` instead."""
    newer_step: asyncio.Event
    next_self_step: Optional[TieredTime]
    """The next self-scheduled step for this simulator."""

    progress: Progress
    """This simulator's progress in mosaik time.

    This simulator has done all its work before time `progress`, so
    other simulator can rely on this simulator's output until this time.
    """
    last_step: TieredTime
    """The most recent step this simulator performed."""
    current_step: Optional[TieredTime]

    output_time: TieredTime  # type: ignore  # set on first get_data
    """The output time associated with `data`. Usually, this will be equal to
    `last_step` but simulators may specify a different time for their output."""
    data: OutputData  # type: ignore  # set on first get_data
    """The newest data returned by this simulator."""
    task: asyncio.Task[None]
    """The asyncio.Task for this simulator."""

    outputs: Optional[Dict[Time, OutputData]]
    tqdm: tqdm.tqdm[NoReturn]  # type: ignore

    def __init__(
        self,
        sid: SimId,
        connection: Proxy,
        depth: int = 1,
    ):
        self.sid = sid
        self._proxy = connection

        self.type = connection.meta["type"]
        self.supports_set_events = connection.meta.get("set_events", False)
        # Simulation state
        self.started = False
        self.last_step = TieredTime(-1, *([0] * (depth - 1)))
        self.current_step = None
        if self.type != "event-based":
            self.next_steps = [TieredTime(*([0] * depth))]
        else:
            self.next_steps = []
        self.next_self_step = None
        self.progress = Progress(TieredTime(*([0] * depth)))

        self.to_world_time = TieredDuration(0, cutoff=1, pre_length=depth)
        self.from_world_time = TieredDuration(*([0] * depth), cutoff=1, pre_length=1)

        self.inputs_from_set_data = {}
        self.persistent_inputs = {}
        self.timed_input_buffer = TimedInputBuffer()

        self.successors_to_wait_for = {}
        self.successors = {}
        self.triggering_ancestors = {}
        self.triggers = {}
        self.output_to_push = {}
        self.pulled_inputs = {}

        self.task = None  # type: ignore  # will be set in World.run
        self.newer_step = asyncio.Event()
        self.is_in_step = False

        self.input_delays = {}

        self.output_request = {}

        self.outputs = None

    def schedule_step(self, tiered_time: TieredTime):
        """Schedule a step for this simulator at the given time. This
        will trigger a re-evaluation whether this simulator's next
        step is settled, provided that the new step is earlier than the
        old one and the simulator is currently awaiting it's next
        settled step.
        """
        if tiered_time in self.next_steps:
            return tiered_time

        is_earlier = not self.next_steps or tiered_time < self.next_steps[0]
        hq.heappush(self.next_steps, tiered_time)
        if is_earlier:
            self.newer_step.set()

    async def setup_done(self):
        return await self._proxy.send(["setup_done", (), {}])

    async def step(
        self, time: Time, inputs: InputData, max_advance: Time
    ) -> Optional[Time]:
        try:
            return await self._proxy.send(["step", (time, inputs, max_advance), {}])
        except TypeError:  # from JSON serialization
            # Find source for more precise error message
            encoder = JSONEncoder()
            error = NonSerializableOutputsError(self.sid)
            for dest_eid, entity_inputs in inputs.items():
                for dest_attr, attr_inputs in entity_inputs.items():
                    for src_id, value in attr_inputs.items():
                        try:
                            encoder.encode(value)
                        except TypeError as e:
                            error.add_error(dest_eid, dest_attr, src_id, e)
            if error:
                raise error
            else:  # no culprits found, raise original exception
                raise

    async def get_data(self, outputs: OutputRequest) -> OutputData:
        return await self._proxy.send(["get_data", (outputs,), {}])

    def get_output_for(self, time: Time) -> OutputData:
        assert self.outputs is not None
        for data_time, value in reversed(self.outputs.items()):
            if data_time <= time:
                return value

        return {}

    async def stop(self):
        """
        Stop the simulator behind the proxy.
        """
        await self._proxy.stop()

    def __repr__(self):
        return f"<{self.__class__.__name__} sid={self.sid!r}>"


class MosaikRemote(mosaik_api_v3.MosaikProxy):
    world: AsyncWorld
    sid: SimId

    def __init__(self, world: AsyncWorld, sid: SimId):
        self.world = world
        self.sid = sid

    @property
    def sim(self):
        return self.world.sims[self.sid]

    async def get_progress(self) -> float:
        """
        Return the current simulation progress from
        :attr:`~mosaik.scenario.World.sim_progress`.
        """
        return self.world.sim_progress

    async def get_related_entities(
        self, entities: Union[FullId, List[FullId], None] = None
    ) -> Union[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        Return information about the related entities of *entities*.

        If *entities* omitted (or ``None``), return the complete entity
        graph, e.g.::

            {
                'nodes': {
                    'sid_0.eid_0': {'type': 'A'},
                    'sid_0.eid_1': {'type': 'B'},
                    'sid_1.eid_0': {'type': 'C'},
                },
                'edges': [
                    ['sid_0.eid_0', 'sid_1.eid0', {}],
                    ['sid_0.eid_1', 'sid_1.eid0', {}],
                ],
            }

        If *entities* is a single string (e.g., ``sid_1.eid_0``), return a dict
        containing all entities related to that entity::

            {
                'sid_0.eid_0': {'type': 'A'},
                'sid_0.eid_1': {'type': 'B'},
            }

        If *entities* is a list of entity IDs (e.g., ``['sid_0.eid_0',
        'sid_0.eid_1']``), return a dict mapping each entity to a dict of
        related entities::

            {
                'sid_0.eid_0': {
                    'sid_1.eid_0': {'type': 'B'},
                },
                'sid_0.eid_1': {
                    'sid_1.eid_1': {'type': 'B'},
                },
            }
        """
        graph = self.world.entity_graph
        if entities is None:
            # repackage NodeViews and EdgeViews to maintain compatibility
            nodes_list = literal_eval(str(graph.nodes(data=True)))
            nodes_dict = {node[0]: node[1] for node in nodes_list}

            edges_list = literal_eval(str(graph.edges))
            edges_tuple = tuple(list(edge) + [{}] for edge in edges_list)

            return {"nodes": nodes_dict, "edges": edges_tuple}
        elif isinstance(entities, str):
            return {n: graph.nodes[n] for n in graph[entities]}
        else:
            return {eid: {n: graph.nodes[n] for n in graph[eid]} for eid in entities}

    async def get_data(self, attrs: Dict[FullId, List[Attr]]) -> Dict[str, Any]:
        """
        .. warning::
            This method is deprecated and will be removed in a future release.
            Implement cyclic data flow using time-shifted and weak connections instead.

        Return the data for the requested attributes *attrs*.

        *attrs* is a dict of (fully qualified) entity IDs mapping to lists
        of attribute names (``{'sid/eid': ['attr1', 'attr2']}``).

        The return value is a dictionary, which maps the input entity IDs to
        data dictionaries, which in turn map attribute names to their
        respective values:
        (``{'sid/eid': {'attr1': val1, 'attr2': val2}}``).
        """
        assert self.sim.is_in_step, "get_data must happen in step"
        assert self.sim.current_step is not None, "no current step time"

        data: Dict[FullId, Dict[Attr, Any]] = {}
        missing: Dict[SimId, OutputRequest] = collections.defaultdict(
            lambda: collections.defaultdict(list)
        )
        # Try to get data from cache
        for full_id, attr_names in attrs.items():
            sid, eid = full_id.split(FULL_ID_SEP, 1)
            src_sim = self.world.sims[sid]
            # Check if async_requests are enabled.
            self._assert_async_requests(src_sim, self.sim)
            if self.world.use_cache:
                cache_slice = src_sim.get_output_for(self.sim.last_step.time)
            else:
                cache_slice = {}

            data[full_id] = {}
            for attr in attr_names:
                try:
                    data[full_id][attr] = cache_slice[eid][attr]
                except KeyError:
                    missing[sid][eid].append(attr)

        # Query simulator for data not in the cache
        for sid, attrs in missing.items():
            dep = self.world.sims[sid]
            # assert dep.progress.value > self.sim.current_step >= dep.last_step, \
            #    "sim progress wrong for async requests"
            dep_data = await dep._proxy.send(["get_data", (attrs,), {}])
            for eid, vals in dep_data.items():
                # Maybe there's already an entry for full_id, so we need
                # to update the dict in that case.
                data.setdefault(FULL_ID % (sid, eid), {}).update(vals)

        return data

    async def set_data(self, data: Dict[FullId, Dict[Attr, Any]]):
        """
        .. warning::
            This method is deprecated and will be removed in a future release.
            Implement cyclic data flow using time-shifted and weak connections instead.

        Set *data* as input data for all affected simulators.

        *data* is a dictionary mapping source entity IDs to destination entity
        IDs with dictionaries of attributes and values (``{'src_full_id':
        {'dest_full_id': {'attr1': 'val1', 'attr2': 'val2'}}}``).
        """
        for src_full_id, dest in data.items():
            for full_id, attributes in dest.items():
                sid, eid = full_id.split(FULL_ID_SEP, 1)
                src_sim = self.world.sims[sid]
                self._assert_async_requests(src_sim, self.sim)
                inputs = src_sim.inputs_from_set_data.setdefault(eid, {})
                for attr, val in attributes.items():
                    inputs.setdefault(attr, {})[src_full_id] = val

    async def set_event(self, event_time: Time):
        """
        Schedules an event/step at simulation time *event_time*.
        """
        sim = self.world.sims[self.sid]
        if not self.world.rt_factor:
            raise SimulationError(
                f"Simulator '{self.sid}' tried to set an event in non-real-time "
                "mode."
            )
        if event_time < self.world.until:
            sim.schedule_step(TieredTime(event_time))
        else:
            logger.warning(
                "Event set at {event_time} by {sim_id} is after simulation end {until} "
                "and will be ignored.",
                event_time=event_time,
                sim_id=sim.sid,
                until=self.world.until,
            )

    def _assert_async_requests(self, src_sim: SimRunner, dest_sim: SimRunner):
        """
        Check if async. requests are allowed from *dest_sid* to *src_sid*
        and raise a :exc:`ScenarioError` if not.
        """
        if dest_sim not in src_sim.successors:
            raise ScenarioError(
                f"No connection from {src_sim.sid} to {dest_sim.sid}: You need to "
                "connect entities from both simulators and set `async_requests=True`."
            )
        if dest_sim not in src_sim.successors_to_wait_for:
            raise ScenarioError(
                f"Async. requests not enabled for the connection from {src_sim.sid} to "
                f"{dest_sim.sid}. Add the argument `async_requests=True` to the "
                f"connection of entities from {src_sim.sid} to {dest_sim.sid}."
            )


class StarterCollection(object):
    """
    This class provides a singleton instance of a collection of simulation
    starters. Default starters are:
    - python: start_inproc
    - cmd: start_proc
    - connect: start_connect

    External packages may add additional methods of starting simulations by
    adding new elements:

        from mosaik.simmanager import StarterCollection
        s = StarterCollection()
        s['my_starter'] = my_starter_func
    """

    # Singleton instance of the starter collection.
    __instance: (
        OrderedDict[str, Callable[..., Coroutine[Any, Any, BaseProxy]]] | None
    ) = None

    def __new__(cls) -> OrderedDict[str, Callable[..., Coroutine[Any, Any, BaseProxy]]]:
        if StarterCollection.__instance is None:
            # Create collection with default starters (i.e., starters defined
            # my mosaik core).
            StarterCollection.__instance = collections.OrderedDict(
                python=start_inproc, cmd=start_proc, connect=start_connect
            )

        return StarterCollection.__instance


class TimedInputBuffer:
    """
    A buffer to store inputs with its corresponding *time*.

    When the data is queried for a specific *step* time, all entries with
    *time* <= *step* are added to the input_dictionary.

    If there are several entries for the same connection at the same time, only
    the most recent value is added.
    """

    input_queue: List[Tuple[Time, int, FullId, EntityId, Attr, Any]]

    def __init__(self):
        self.input_queue = []
        self.counter = itertools.count()  # Used to chronologically sort entries

    def add(
        self,
        time: Time,
        src_sid: SimId,
        src_eid: EntityId,
        dest_eid: EntityId,
        dest_attr: Attr,
        value: Any,
    ):
        src_full_id = f"{src_sid}.{src_eid}"
        hq.heappush(
            self.input_queue,
            (time, next(self.counter), src_full_id, dest_eid, dest_attr, value),
        )

    def get_input(self, input_dict: InputData, step: Time) -> InputData:
        while len(self.input_queue) > 0 and self.input_queue[0][0] <= step:
            _, _, src_full_id, eid, attr, value = hq.heappop(self.input_queue)
            input_dict.setdefault(eid, {}).setdefault(attr, {})[src_full_id] = value

        return input_dict

    def __bool__(self):
        return bool(len(self.input_queue))
